# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
It provides CO2MPAS model `dsp` to predict light-vehicles' CO2 emissions.

Docstrings should provide sufficient understanding for any individual function.

Sub-Modules:

.. currentmodule:: co2mpas.core.model.physical

.. autosummary::
    :nosignatures:
    :toctree: physical/

    vehicle
    wheels
    final_drive
    gear_box
    clutch_tc
    cycle
    electrics
    engine
    control
    after_treat
    co2
"""

import schedula as sh
from .cycle import dsp as _cycle
from .vehicle import dsp as _vehicle
from .wheels import dsp as _wheels
from .final_drive import dsp as _final_drive
from .gear_box import dsp as _gear_box
from .clutch_tc import dsp as _clutch_torque_converter
from .electrics import dsp as _electrics
from .engine import dsp as _engine
from .control import dsp as _control
from .after_treat import dsp as _after_treat
from .co2 import dsp as _co2

dsp = sh.BlueDispatcher(
    name='CO2MPAS physical model',
    description='Wraps all functions needed to calibrate and predict '
                'light-vehicles\' CO2 emissions.'
)

dsp.add_data('path_elevations', wildcard=True)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='cycle_model',
    dsp=_cycle,
    inputs=(
        'max_speed_velocity_ratio', 'engine_speed_at_max_power', 'unladen_mass',
        'downscale_factor_threshold', 'speed_velocity_ratios', 'k1', 'k2', 'k5',
        'idle_engine_speed', 'downscale_phases', 'engine_max_power', 'max_gear',
        'engine_max_speed', 'full_load_curve', 'accelerations', 'motive_powers',
        'gear_box_type', 'road_loads', 'cycle_type', 'wltp_class', 'bag_phases',
        'downscale_factor', 'inertial_factor', 'vehicle_mass', 'times', 'gears',
        'time_sample_frequency', 'wltp_base_model', 'max_velocity', 'max_time',
        'velocities',
    ),
    outputs=(
        'initial_temperature', 'phases_integration_times', 'times', 'gears',
        'velocities', 'theoretical_motive_powers', 'theoretical_velocities'
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='vehicle_model',
    dsp=_vehicle,
    inputs=(
        'traction_acceleration_limit', 'traction_deceleration_limit', 'n_wheel',
        'aerodynamic_drag_coefficient', 'rolling_resistance_coeff', 'fuel_mass',
        'wheel_drive_load_fraction', 'n_wheel_drive', 'obd_velocities', 'times',
        'static_friction', 'initial_velocity', 'tyre_category', 'vehicle_width',
        'vehicle_body', 'road_state', 'cargo_mass', 'angle_slope', 'tyre_class',
        'n_passengers', 'passenger_mass', 'air_density', 'f0_uncorrected', 'f1',
        'n_dyno_axes', 'velocities', 'angle_slopes', 'correct_f0', 'cycle_type',
        'curb_mass', 'elevations', 'frontal_area', 'has_roof_box', 'road_loads',
        'unladen_mass', 'inertial_factor', 'vehicle_height', 'vehicle_category',
        'tyre_state', 'vehicle_mass', 'f0', 'path_distances', 'path_elevations',
        'wheel_powers', 'f2', 'test_mass', 'vehicle_mass_running_order',
        'aerodynamic_drag_area'
    ),
    outputs=(
        'wheel_drive_load_fraction', 'traction_acceleration_limit', 'curb_mass',
        'traction_deceleration_limit', 'inertial_factor', 'motive_powers', 'f0',
        'static_friction', 'accelerations', 'angle_slopes', 'n_dyno_axes', 'f1',
        'unladen_mass', 'vehicle_mass', 'velocities', 'road_loads', 'distances',
        'f2', 'test_mass', 'aerodynamic_drag_area'
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='wheels_model',
    dsp=_wheels,
    inputs=(
        'final_drive_ratios', 'gear_box_ratios', 'gears', 'velocities', 'times',
        'tyre_dynamic_rolling_coefficient', 'tyre_code', 'plateau_acceleration',
        'gear_box_speeds_in', 'idle_engine_speed', 'stop_velocity', 'on_engine',
        'engine_speeds_out', 'tyre_dimensions', 'accelerations', 'wheel_powers',
        'velocity_speed_ratios', 'front_tyre_code', 'motive_powers', 'r_wheels',
        'change_gear_window_width', 'rear_tyre_code', 'wheel_drive', 'r_dynamic'
    ),
    outputs=(
        'r_dynamic', 'r_wheels', 'tyre_code', 'wheel_powers', 'wheel_speeds',
        'tyre_dynamic_rolling_coefficient', 'wheel_torques'
    ),
    inp_weight={'r_dynamic': 3}
)


@sh.add_function(dsp, outputs=['final_drive_powers_out'])
def calculate_final_drive_powers_out(
        wheel_powers, motor_p4_front_powers, motor_p4_rear_powers):
    """
    Calculate final drive power out [kW].

    :param wheel_powers:
        Power at the wheels [kW].
    :type wheel_powers: numpy.array | float

    :param motor_p4_front_powers:
        Power at motor P4 front [kW].
    :type motor_p4_front_powers: numpy.array | float

    :param motor_p4_rear_powers:
        Power at motor P4 rear [kW].
    :type motor_p4_rear_powers: numpy.array | float

    :return:
        Final drive power out [kW].
    :rtype: numpy.array | float
    """
    return wheel_powers - motor_p4_front_powers - motor_p4_rear_powers


@sh.add_function(dsp, outputs=['wheel_powers'])
def calculate_wheel_powers(
        final_drive_powers_out, motor_p4_front_powers, motor_p4_rear_powers):
    """
    Calculate the power at the wheels [kW].

    :param final_drive_powers_out:
        Final drive power out [kW].
    :type final_drive_powers_out: numpy.array | float

    :param motor_p4_front_powers:
        Power at motor P4 front [kW].
    :type motor_p4_front_powers: numpy.array | float

    :param motor_p4_rear_powers:
        Power at motor P4 rear [kW].
    :type motor_p4_rear_powers: numpy.array | float

    :return:
        Power at the wheels [kW].
    :rtype: numpy.array | float
    """
    return final_drive_powers_out + motor_p4_front_powers + motor_p4_rear_powers


dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='final_drive_model',
    dsp=_final_drive,
    inputs=(
        'final_drive_efficiency', 'final_drive_ratio', 'n_wheel_drive', 'gears',
        'final_drive_powers_out', 'final_drive_powers_in', 'final_drive_ratios',
        'gear_box_type', 'n_gears', 'wheel_drive',
        {'wheel_speeds': 'final_drive_speeds_out'}
    ),
    outputs=(
        'final_drive_torques_in', 'final_drive_ratios', 'final_drive_speeds_in',
        'final_drive_mean_efficiency', 'final_drive_powers_out',
        'final_drive_powers_in', 'n_wheel_drive'
    )
)


@sh.add_function(dsp, outputs=['gear_box_powers_out'])
def calculate_gear_box_powers_out(
        final_drive_powers_in, motor_p3_front_powers, motor_p3_rear_powers):
    """
    Calculate gear box power vector [kW].

    :param final_drive_powers_in:
        Final drive power in [kW].
    :type final_drive_powers_in: numpy.array | float

    :param motor_p3_front_powers:
        Power at motor P3 front [kW].
    :type motor_p3_front_powers: numpy.array | float

    :param motor_p3_rear_powers:
        Power at motor P3 rear [kW].
    :type motor_p3_rear_powers: numpy.array | float

    :return:
        Gear box power vector [kW].
    :rtype: numpy.array | float
    """
    return final_drive_powers_in - motor_p3_front_powers - motor_p3_rear_powers


@sh.add_function(dsp, outputs=['final_drive_powers_in'])
def calculate_final_drive_powers_in(
        gear_box_powers_out, motor_p3_front_powers, motor_p3_rear_powers):
    """
    Calculate final drive power in [kW].

    :param gear_box_powers_out:
        Gear box power vector [kW].
    :type gear_box_powers_out: numpy.array | float

    :param motor_p3_front_powers:
        Power at motor P3 front [kW].
    :type motor_p3_front_powers: numpy.array | float

    :param motor_p3_rear_powers:
        Power at motor P3 rear [kW].
    :type motor_p3_rear_powers: numpy.array | float

    :return:
        Final drive power in [kW].
    :rtype: numpy.array | float
    """
    return gear_box_powers_out + motor_p3_front_powers + motor_p3_rear_powers


dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='gear_box_model',
    dsp=_gear_box,
    inputs=(
        'has_gear_box_thermal_management', 'has_torque_converter', 'velocities',
        'maximum_vehicle_laden_mass', 'maximum_velocity', 'min_engine_on_speed',
        'engine_temperatures', 'engine_mass', 'engine_max_power', 'MVL',
        'initial_gear_box_temperature', 'gear_box_temperature_references', 'f0',
        'time_cold_hot_transition', 'specific_gear_shifting', 'full_load_curve',
        'gear_box_speeds_in', 'engine_speeds_out', 'gear_box_type', 'on_engine',
        'CMV', 'CMV_Cold_Hot', 'engine_max_torque', 'engine_speed_at_max_power',
        'gear_box_powers_in', 'CVT', 'idle_engine_speed', 'cycle_type', 'times',
        'engine_thermostat_temperature', 'last_gear_box_ratio', 'stop_velocity',
        'engine_speed_at_max_velocity', 'first_gear_box_ratio', 'motive_powers',
        'change_gear_window_width', 'gear_box_powers_out', 'final_drive_ratios',
        'gear_box_efficiency_parameters_cold_hot', 'GSPV_Cold_Hot', 'r_dynamic',
        'final_drive_mean_efficiency', 'accelerations', 'velocity_speed_ratios',
        'fuel_saving_at_strategy', 'use_dt_gear_shifting', 'road_loads', 'DTGS',
        'gear_box_efficiency_constants', 'engine_max_speed', 'n_gears', 'gears',
        'plateau_acceleration', 'full_load_speeds', 'gear_box_ratios', 'GSPV',
        'max_velocity_full_load_correction',
        {
            'final_drive_speeds_in': 'gear_box_speeds_out',
            'initial_engine_temperature': 'initial_gear_box_temperature',
            'initial_temperature': 'initial_gear_box_temperature'
        }
    ),
    outputs=(
        'gear_box_ratios', 'gear_box_speeds_in', 'gear_box_temperatures', 'MVL',
        'equivalent_gear_box_heat_capacity', 'max_speed_velocity_ratio', 'GSPV',
        'gear_box_efficiencies', 'speed_velocity_ratios', 'last_gear_box_ratio',
        'gear_box_mean_efficiency_guess', 'gear_box_mean_efficiency', 'n_gears',
        'engine_speed_at_max_velocity', 'first_gear_box_ratio', 'GSPV_Cold_Hot',
        'specific_gear_shifting', 'velocity_speed_ratios', 'gear_box_powers_in',
        'gear_box_powers_out', 'maximum_velocity', 'CMV_Cold_Hot', 'CMV', 'CVT',
        'gear_box_torques_in', 'gear_shifts', 'gears', 'max_gear', 'DTGS',
    ),
    inp_weight={'initial_temperature': 5}
)


@sh.add_function(dsp, outputs=['clutch_tc_powers_out'])
def calculate_clutch_tc_powers_out(gear_box_powers_in, motor_p2_powers):
    """
    Calculate gear box power vector [kW].

    :param gear_box_powers_in:
        Gear box power vector [kW].
    :type gear_box_powers_in: numpy.array | float

    :param motor_p2_powers:
        Power at motor P2 [kW].
    :type motor_p2_powers: numpy.array | float

    :return:
        Clutch or torque converter power out [kW].
    :rtype: numpy.array | float
    """
    return gear_box_powers_in - motor_p2_powers


@sh.add_function(dsp, outputs=['gear_box_powers_in'])
def calculate_gear_box_powers_in(clutch_tc_powers_out, motor_p2_powers):
    """
    Calculate clutch or torque converter power out [kW].

    :param clutch_tc_powers_out:
        Clutch or torque converter power out [kW].
    :type clutch_tc_powers_out: numpy.array | float

    :param motor_p2_powers:
        Power at motor P2 [kW].
    :type motor_p2_powers: numpy.array | float

    :return:
        Gear box power vector [kW].
    :rtype: numpy.array | float
    """
    return clutch_tc_powers_out + motor_p2_powers


dsp.add_dispatcher(
    include_defaults=True,
    dsp=_clutch_torque_converter,
    dsp_id='clutch_torque_converter_model',
    inputs=(
        'engine_speeds_base', 'gear_box_torques_in', 'm1000_curve_norm_torques',
        'engine_speeds_out', 'stand_still_torque_ratio', 'has_torque_converter',
        'engine_speeds_out_hot', 'm1000_curve_ratios', 'clutch_window', 'gears',
        'torque_converter_speed_model', 'm1000_curve_factor', 'full_load_curve',
        'engine_max_speed', 'idle_engine_speed', 'accelerations', 'gear_shifts',
        'lockup_speed_ratio', 'clutch_speed_model', 'clutch_tc_powers', 'times',
        'clutch_tc_powers_out', 'gear_box_speeds_in', 'stop_velocity',
        'gear_box_type', 'velocities',
    ),
    outputs=(
        'clutch_tc_powers_out', 'clutch_phases', 'clutch_tc_mean_efficiency',
        'clutch_window', 'clutch_tc_speeds_delta', 'has_torque_converter',
        'lockup_speed_ratio', 'stand_still_torque_ratio', 'm1000_curve_factor',
        'torque_converter_speed_model', 'clutch_tc_powers', 'clutch_speed_model'
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='electric_model',
    dsp=_electrics,
    inputs=(
        'motor_p0_efficiency', 'motor_p0_maximum_power', 'motor_p0_speed_ratio',
        'motor_p0_electric_powers', 'motor_p0_torques', 'motor_p0_rated_speed',
        'motor_p0_powers', 'motor_p0_maximum_torque', 'motor_p0_speeds',
        'motor_p1_efficiency', 'motor_p1_maximum_power', 'motor_p1_speed_ratio',
        'motor_p1_electric_powers', 'motor_p1_torques', 'motor_p1_rated_speed',
        'motor_p1_powers', 'motor_p1_maximum_torque', 'motor_p1_speeds',
        'motor_p2_planetary_efficiency', 'motor_p2_planetary_maximum_power',
        'motor_p2_planetary_speed_ratio', 'motor_p2_planetary_electric_powers',
        'motor_p2_planetary_torques', 'motor_p2_planetary_rated_speed',
        'motor_p2_planetary_powers', 'motor_p2_planetary_maximum_torque',
        'motor_p2_planetary_speeds',
        'motor_p2_efficiency', 'motor_p2_maximum_power', 'motor_p2_speed_ratio',
        'motor_p2_electric_powers', 'motor_p2_torques', 'motor_p2_rated_speed',
        'motor_p2_powers', 'motor_p2_maximum_torque', 'motor_p2_speeds',
        'motor_p3_front_efficiency', 'motor_p3_front_maximum_power',
        'motor_p3_front_speed_ratio', 'motor_p3_front_electric_powers',
        'motor_p3_front_torques', 'motor_p3_front_rated_speed',
        'motor_p3_front_powers', 'motor_p3_front_maximum_torque',
        'motor_p3_front_speeds',
        'motor_p3_rear_efficiency', 'motor_p3_rear_maximum_power',
        'motor_p3_rear_speed_ratio', 'motor_p3_rear_electric_powers',
        'motor_p3_rear_torques', 'motor_p3_rear_rated_speed',
        'motor_p3_rear_powers', 'motor_p3_rear_maximum_torque',
        'motor_p3_rear_speeds',
        'motor_p4_front_efficiency', 'motor_p4_front_maximum_power',
        'motor_p4_front_speed_ratio', 'motor_p4_front_electric_powers',
        'motor_p4_front_torques', 'motor_p4_front_rated_speed',
        'motor_p4_front_powers', 'motor_p4_front_maximum_torque',
        'motor_p4_front_speeds',
        'motor_p4_rear_efficiency', 'motor_p4_rear_maximum_power',
        'motor_p4_rear_speed_ratio', 'motor_p4_rear_electric_powers',
        'motor_p4_rear_torques', 'motor_p4_rear_rated_speed',
        'motor_p4_rear_powers', 'motor_p4_rear_maximum_torque',
        'motor_p4_rear_speeds',
        'has_motor_p0', 'has_motor_p1', 'has_motor_p2_planetary',
        'has_motor_p2', 'has_motor_p3_front', 'has_motor_p3_rear',
        'has_motor_p4_front', 'has_motor_p4_rear',
        'service_battery_initialization_time', 'engine_starts', 'accelerations',
        'starter_efficiency', 'delta_time_engine_starter', 'gear_box_speeds_in',
        'service_battery_charging_statuses', 'alternator_currents', 'on_engine',
        'starter_nominal_voltage', 'engine_speeds_out', 'alternator_efficiency',
        'service_battery_state_of_charges', 'alternator_current_model', 'times',
        'alternator_charging_currents', 'engine_moment_inertia', 'wheel_speeds',
        'drive_battery_ocv', 'drive_battery_r0', 'drive_battery_n_series_cells',
        'initial_service_battery_state_of_charge', 'alternator_electric_powers',
        'service_battery_start_window_width', 'service_battery_nominal_voltage',
        'cycle_type', 'service_battery_electric_powers', 'service_battery_load',
        'initial_drive_battery_state_of_charge', 'service_battery_status_model',
        'final_drive_speeds_in', 'dcdc_charging_currents', 'dcdc_current_model',
        'service_battery_electric_powers_supply_threshold', 'engine_powers_out',
        'service_battery_state_of_charge_balance_window', 'drive_battery_loads',
        'drive_battery_currents', 'service_battery_loads', 'drive_battery_load',
        'service_battery_state_of_charge_balance', 'alternator_nominal_voltage',
        'dcdc_converter_efficiency', 'has_energy_recuperation', 'motive_powers',
        'starter_electric_powers', 'dcdc_converter_currents', 'planetary_ratio',
        'dcdc_converter_electric_powers_demand', 'planetary_mean_efficiency',
        'electrical_hybridization_degree', 'drive_battery_n_parallel_cells',
        'drive_battery_state_of_charges', 'drive_battery_electric_powers',
        'dcdc_converter_electric_powers', 'drive_battery_nominal_voltage',
        'service_battery_capacity_kwh', 'drive_battery_capacity_kwh',
        'service_battery_currents', 'drive_battery_technology',
        'service_battery_capacity', 'drive_battery_capacity',
        'drive_battery_voltages', 'drive_battery_n_cells',
    ),
    outputs=(
        'motor_p0_electric_powers', 'motor_p0_maximum_power', 'motor_p0_powers',
        'motor_p0_maximum_torque', 'motor_p0_speed_ratio', 'motor_p0_torques',
        'motor_p0_maximum_powers', 'motor_p0_rated_speed', 'motor_p0_speeds',
        'motor_p0_maximum_power_function', 'motor_p1_maximum_power_function',
        'motor_p1_electric_powers', 'motor_p1_maximum_power', 'motor_p1_powers',
        'motor_p1_maximum_torque', 'motor_p1_speed_ratio', 'motor_p1_torques',
        'motor_p1_maximum_powers', 'motor_p1_rated_speed', 'motor_p1_speeds',
        'motor_p2_planetary_maximum_power_function',
        'motor_p2_planetary_electric_powers',
        'motor_p2_planetary_maximum_power', 'motor_p2_planetary_powers',
        'motor_p2_planetary_maximum_torque', 'motor_p2_planetary_speed_ratio',
        'motor_p2_planetary_torques', 'motor_p2_planetary_maximum_powers',
        'motor_p2_planetary_rated_speed', 'motor_p2_planetary_speeds',
        'motor_p2_electric_powers', 'motor_p2_maximum_power', 'motor_p2_powers',
        'motor_p2_maximum_torque', 'motor_p2_speed_ratio', 'motor_p2_torques',
        'motor_p2_maximum_powers', 'motor_p2_rated_speed', 'motor_p2_speeds',
        'motor_p3_front_electric_powers', 'motor_p3_front_maximum_power',
        'motor_p3_front_powers', 'motor_p3_front_maximum_torque',
        'motor_p3_front_speed_ratio', 'motor_p3_front_torques',
        'motor_p3_front_maximum_powers', 'motor_p3_front_rated_speed',
        'motor_p3_front_speeds',
        'motor_p3_rear_electric_powers', 'motor_p3_rear_maximum_power',
        'motor_p3_rear_powers', 'motor_p3_rear_maximum_torque',
        'motor_p3_rear_speed_ratio', 'motor_p3_rear_torques',
        'motor_p3_rear_maximum_powers', 'motor_p3_rear_rated_speed',
        'motor_p3_rear_speeds',
        'motor_p4_front_electric_powers', 'motor_p4_front_maximum_power',
        'motor_p4_front_powers', 'motor_p4_front_maximum_torque',
        'motor_p4_front_speed_ratio', 'motor_p4_front_torques',
        'motor_p4_front_maximum_powers', 'motor_p4_front_rated_speed',
        'motor_p4_front_speeds',
        'motor_p4_rear_electric_powers', 'motor_p4_rear_maximum_power',
        'motor_p4_rear_powers', 'motor_p4_rear_maximum_torque',
        'motor_p4_rear_speed_ratio', 'motor_p4_rear_torques',
        'motor_p4_rear_maximum_powers', 'motor_p4_rear_rated_speed',
        'motor_p4_rear_speeds',
        'has_motor_p0', 'has_motor_p1', 'has_motor_p2_planetary',
        'has_motor_p2', 'has_motor_p3_front', 'has_motor_p3_rear',
        'has_motor_p4_front', 'has_motor_p4_rear', 'is_hybrid',
        'starter_electric_powers', 'final_drive_speeds_in', 'alternator_powers',
        'alternator_electric_powers', 'engine_speeds_out', 'gear_box_speeds_in',
        'delta_time_engine_starter', 'alternator_current_model', 'wheel_speeds',
        'alternator_currents', 'starter_powers', 'starter_model',
        'service_battery_electric_powers_supply_threshold', 'drive_battery_ocv',
        'service_battery_state_of_charge_balance_window', 'drive_battery_model',
        'drive_battery_state_of_charges', 'drive_battery_delta_state_of_charge',
        'initial_drive_battery_state_of_charge', 'drive_battery_n_series_cells',
        'dcdc_converter_currents', 'drive_battery_capacity', 'drive_battery_r0',
        'service_battery_delta_state_of_charge', 'service_battery_status_model',
        'service_battery_state_of_charges', 'service_battery_charging_statuses',
        'dcdc_current_model', 'drive_battery_loads', 'service_battery_currents',
        'starter_currents', 'service_battery_load', 'dcdc_converter_efficiency',
        'drive_battery_voltages', 'service_battery_loads', 'drive_battery_load',
        'service_battery_state_of_charge_balance', 'alternator_nominal_voltage',
        'planetary_mean_efficiency', 'drive_battery_n_cells', 'planetary_ratio',
        'service_battery_initialization_time', 'drive_battery_n_parallel_cells',
        'initial_service_battery_state_of_charge', 'service_battery_capacity',
        'service_battery_electric_powers', 'service_battery_nominal_voltage',
        'dcdc_converter_electric_powers_demand', 'starter_nominal_voltage',
        'drive_battery_electric_powers', 'drive_battery_capacity_kwh',
        'drive_battery_nominal_voltage', 'motors_electric_powers',
        'drive_battery_currents', 'service_battery_capacity_kwh'
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='engine_model',
    dsp=_engine,
    inputs=(
        'engine_thermostat_temperature', 'engine_thermostat_temperature_window',
        'engine_powers_out', 'engine_speeds_out', 'stop_velocity', 'velocities',
        'initial_friction_params', 'min_engine_on_speed', 'fuel_carbon_content',
        'motor_p0_powers', 'motor_p1_powers', 'auxiliaries_torque_loss_factors',
        'idle_engine_speed', 'engine_coolant_temperatures', 'full_load_torques',
        'has_selective_catalytic_reduction', 'engine_has_cylinder_deactivation',
        'engine_has_variable_valve_actuation', 'max_engine_coolant_temperature',
        'engine_fuel_lower_heating_value', 'fuel_consumptions', 'gear_box_type',
        'engine_capacity', 'clutch_tc_speeds_delta', 'phases_integration_times',
        'engine_speed_at_max_power', 'obd_fuel_type_code', 'calibration_status',
        'co2_normalization_references', 'auxiliaries_torque_loss', 'co2_params',
        'idle_engine_speed_median', 'auxiliaries_power_loss', 'engine_is_turbo',
        'engine_max_power', 'clutch_tc_powers', 'full_load_speeds', 'is_hybrid',
        'engine_temperature_regression_model', 'has_lean_burn', 'accelerations',
        'full_load_powers', 'belt_efficiency', 'engine_stroke', 'ignition_type',
        'engine_speeds_out_hot', 'idle_engine_speed_std', 'gear_box_powers_out',
        'engine_n_cylinders', 'engine_max_speed', 'is_cycle_hot', 'engine_type',
        'has_exhausted_gas_recirculation', 'co2_params_calibrated', 'on_engine',
        'engine_idle_fuel_consumption', 'after_treatment_speeds_delta', 'times',
        'active_cylinder_ratios', 'alternator_powers', 'engine_mass', 'on_idle',
        'initial_engine_temperature', 'motor_p2_planetary_powers', 'fuel_type',
        'gross_engine_powers_out', 'phases_co2_emissions', 'engine_max_torque',
        'after_treatment_warm_up_phases', 'phases_distances', 'fuel_density',
        'fuel_consumptions_liters', 'temperature_shift', {
            'initial_temperature': 'initial_engine_temperature'
        }
    ),
    outputs=(
        'identified_co2_emissions', 'idle_engine_speed', 'has_sufficient_power',
        'idle_engine_speed_median', 'idle_engine_speed_std', 'full_load_powers',
        'engine_temperature_regression_model', 'engine_speeds_out', 'fuel_type',
        'engine_thermostat_temperature', 'engine_thermostat_temperature_window',
        'co2_params_calibrated', 'co2_params_initial_guess', 'active_cylinders',
        'initial_friction_params', 'engine_moment_inertia', 'engine_powers_out',
        'extended_phases_co2_emissions', 'gross_engine_powers_out', 'is_hybrid',
        'engine_inertia_powers_losses', 'fuel_consumptions', 'engine_max_speed',
        'engine_speed_at_max_power', 'belt_mean_efficiency', 'engine_max_power',
        'co2_rescaling_scores', 'auxiliaries_power_loss', 'co2_emissions_model',
        'auxiliaries_torque_losses', 'engine_heat_capacity', 'full_load_speeds',
        'active_exhausted_gas_recirculations', 'engine_temperature_derivatives',
        'initial_engine_temperature', 'engine_coolant_temperatures', 'fuel_map',
        'auxiliaries_torque_loss', 'auxiliaries_power_losses', 'missing_powers',
        'extended_phases_integration_times', 'mean_piston_speeds', 'fmep_model',
        'has_exhausted_gas_recirculation', 'active_lean_burns', 'co2_emissions',
        'engine_fuel_lower_heating_value', 'calibration_status', 'brake_powers',
        'engine_idle_fuel_consumption', 'active_variable_valves', 'engine_mass',
        'max_engine_coolant_temperature', 'full_load_curve', 'clutch_tc_powers',
        'engine_speeds_out_hot', 'fuel_carbon_content', 'engine_temperatures',
        'fuel_consumptions_liters', 'engine_max_torque', 'ignition_type',
        'engine_type', 'temperature_shift'
    ),
    inp_weight={'initial_temperature': 5}
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='control_model',
    dsp=_control,
    inputs=(
        'full_load_curve', 'gear_box_speeds_in', 'motor_p0_efficiency', 'gears',
        'motor_p3_front_maximum_power', 'motor_p3_rear_maximum_power', 'ecms_s',
        'final_drive_mean_efficiency', 'gear_box_mean_efficiency', 'velocities',
        'start_stop_hybrid_params', 'motor_p2_maximum_powers', 'has_start_stop',
        'drive_battery_state_of_charges', 'is_hybrid', 'motor_p2_maximum_power',
        'motor_p0_speed_ratio', 'motor_p2_efficiency', 'auxiliaries_power_loss',
        'start_stop_activation_time', 'motor_p0_maximum_power', 'starter_model',
        'is_cycle_hot', 'motor_p2_electric_powers', 'clutch_tc_mean_efficiency',
        'motor_p3_rear_efficiency', 'motor_p0_electric_powers', 'accelerations',
        'min_time_engine_on_after_start', 'motors_electric_powers', 'is_serial',
        'motor_p1_electric_powers', 'dcdc_converter_efficiency', 'hybrid_modes',
        'motor_p1_maximum_power_function', 'motor_p3_front_efficiency', 'times',
        'motor_p3_front_maximum_powers', 'auxiliaries_torque_loss', 'on_engine',
        'motor_p1_maximum_power', 'belt_mean_efficiency', 'drive_battery_model',
        'motor_p4_front_maximum_power', 'motor_p1_speed_ratio', 'motive_powers',
        'motor_p3_rear_maximum_powers', 'motor_p4_front_efficiency', 'fuel_map',
        'motor_p4_rear_efficiency', 'engine_speeds_out_hot', 'start_stop_model',
        'motor_p4_rear_maximum_power', 'engine_powers_out', 'idle_engine_speed',
        'correct_start_stop_with_gears', 'engine_speeds_out', 'planetary_ratio',
        'engine_moment_inertia', 'engine_temperatures', 'final_drive_speeds_in',
        'motor_p2_planetary_electric_powers', 'motor_p2_planetary_speed_ratio',
        'after_treatment_cooling_duration', 'after_treatment_warm_up_duration',
        'planetary_mean_efficiency', 'has_motor_p2_planetary', 'gear_box_type',
        'motor_p2_planetary_maximum_power', 'motor_p4_front_electric_powers',
        'motor_p0_maximum_power_function', 'motor_p3_front_electric_powers',
        'motor_p2_planetary_maximum_power_function', 'motor_p1_efficiency',
        'after_treatment_warm_up_phases', 'motor_p2_planetary_efficiency',
        'gear_box_mean_efficiency_guess', 'motor_p3_rear_electric_powers',
        'motor_p4_rear_electric_powers', 'engine_thermostat_temperature',
        'motor_p4_front_maximum_powers', 'motor_p4_rear_maximum_powers',
        'min_engine_on_speed',
    ),
    outputs=(
        'correct_start_stop_with_gears', 'start_stop_activation_time', 'ecms_s',
        'motor_p0_electric_powers', 'motor_p1_electric_powers', 'engine_starts',
        'engine_speeds_out_hot', 'start_stop_hybrid_params', 'start_stop_model',
        'motor_p3_front_electric_powers', 'engine_speeds_base', 'hybrid_modes',
        'motor_p2_planetary_electric_powers', 'motor_p3_rear_electric_powers',
        'motor_p4_front_electric_powers', 'after_treatment_warm_up_phases',
        'motor_p4_rear_electric_powers', 'motor_p2_electric_powers',
        'force_on_engine', 'on_engine',
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='after_treat_model',
    dsp=_after_treat,
    inputs=(
        'after_treatment_warm_up_phases', 'engine_speeds_out_hot', 'velocities',
        'after_treatment_speed_model', 'engine_speeds_out', 'engine_powers_out',
        'min_engine_on_speed', 'gear_box_speeds_in', 'stop_velocity', 'on_idle',
        'after_treatment_cooling_duration', 'after_treatment_warm_up_duration',
        'after_treatment_speeds_delta', 'after_treatment_power_model', 'times',
        'engine_speeds_base', 'idle_engine_speed', 'is_cycle_hot', 'on_engine',
        'gears', 'is_hybrid', 'engine_starts'
    ),
    outputs=(
        'after_treatment_power_model', 'after_treatment_speed_model', 'on_idle',
        'after_treatment_warm_up_duration', 'after_treatment_cooling_duration',
        'after_treatment_warm_up_phases', 'after_treatment_speeds_delta',
        'engine_speeds_base',
    )
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='co2_model',
    dsp=_co2,
    inputs=(
        'co2_emission_extra_high', 'phases_co2_emissions', 'mean_piston_speeds',
        'co2_params_calibrated', 'enable_phases_willans', 'co2_emission_medium',
        'engine_fuel_lower_heating_value', 'co2_emission_low', 'missing_powers',
        'co2_emission_value', 'ki_multiplicative', 'co2_emission_high', 'times',
        'kco2_nedc_correction_factor', 'phases_integration_times', 'fmep_model',
        'engine_speeds_out', 'co2_emission_UDC', 'fuel_density', 'angle_slopes',
        'has_periodically_regenerating_systems', 'motive_powers', 'ki_additive',
        'min_engine_on_speed', 'fuel_carbon_content', 'velocities', 'fuel_type',
        'fuel_carbon_content_percentage', 'engine_powers_out', 'enable_willans',
        'speed_distance_correction', 'theoretical_velocities', 'rcb_correction',
        'theoretical_motive_powers', 'alternator_efficiency', 'engine_capacity',
        'co2_emission_EUDC', 'engine_max_power', 'engine_stroke', 'engine_type',
        'atct_family_correction_factor', 'cycle_type', 'distances', 'is_hybrid',
        'drive_battery_delta_state_of_charge', 'after_treatment_warm_up_phases',
        'drive_battery_electric_powers', 'engine_temperatures', 'accelerations',
        'drive_battery_nominal_voltage', 'drive_battery_capacity', 'is_plugin',
        'service_battery_electric_powers', 'force_on_engine', 'co2_emissions',
        'kco2_wltp_correction_factor', 'fuel_consumptions_liters',
        'input_type'
    ),
    outputs=(
        'phases_willans_factors', 'phases_co2_emissions', 'fuel_carbon_content',
        'fuel_carbon_content_percentage', 'phases_distances', 'willans_factors',
        'declared_sustaining_co2_emission_value', 'kco2_wltp_correction_factor',
        'kco2_nedc_correction_factor', 'co2_emission_high', 'co2_emission_EUDC',
        'phases_fuel_consumptions', 'co2_emission_medium', 'optimal_efficiency',
        'declared_co2_emission_value', 'co2_emission_value', 'co2_emission_low',
        'ki_multiplicative', 'phases_indices', 'fuel_consumptions_liters_value',
        'corrected_co2_emission_value', 'fuel_heating_value', 'rcb_correction',
        'engine_fuel_lower_heating_value', 'fuel_density', 'co2_emission_UDC',
        'corrected_sustaining_co2_emission_value', 'co2_emission_extra_high',
        'is_plugin'
    )
)

try:
    from co2mpas_driver.co2mpas import plugin_physycal

    dsp = plugin_physycal(dsp)
except ImportError:
    pass



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and constants to define the after_treatment_model selector.
"""
import schedula as sh
from ._core import define_sub_model
from ...physical.after_treat import dsp as _after_treat

#: Model name.
name = 'after_treatment_model'

#: Parameters that constitute the model.
models = ['after_treatment_speed_model', 'after_treatment_warm_up_duration',
          'after_treatment_cooling_duration']

#: Inputs required to run the model.
inputs = [
    'times', 'on_engine', 'engine_speeds_out_hot', 'is_cycle_hot', 'on_idle',
    'is_hybrid'
]

#: Relevant outputs of the model.
outputs = ['engine_speeds_base']

#: Targets to compare the outputs of the model.
targets = outputs

#: Extra inputs for the metrics.
metrics_inputs = ['after_treatment_warm_up_phases']


def metric_after_treatment_speed_model(
        y_true, y_pred, after_treatment_warm_up_phases):
    """
    Metric for the `after_treatment_speed_model`.

    :param y_true:
        Reference engine speed vector [RPM].
    :type y_true: numpy.array

    :param y_pred:
        Predicted engine speed vector [RPM].
    :type y_pred: numpy.array

    :param after_treatment_warm_up_phases:
        Phases when engine speed is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :return:
        Error.
    :rtype: float
    """
    b = after_treatment_warm_up_phases
    if b.any():
        from co2mpas.utils import mae
        return float(mae(y_true[b], y_pred[b]))
    else:
        return 0


#: Metrics to compare outputs with targets.
metrics = sh.map_list(targets, metric_after_treatment_speed_model)

#: Upper score limits to raise the warnings.
up_limit = sh.map_list(targets, 160)

#: Prediction model.
# noinspection PyProtectedMember
dsp = sh.Blueprint(_after_treat, inputs, outputs, models)._set_cls(
    define_sub_model
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and constants to define the engine_speed_model selector.
"""
import schedula as sh
from ._core import define_sub_model
from ...physical import dsp as _physical
from ...physical.clutch_tc import calculate_clutch_phases

#: Model name.
name = 'engine_speed_model'

#: Parameters that constitute the model.
models = [
    'final_drive_ratios', 'gear_box_ratios', 'idle_engine_speed_median',
    'idle_engine_speed_std', 'CVT', 'max_speed_velocity_ratio',
    'tyre_dynamic_rolling_coefficient'
]

#: Inputs required to run the model.
inputs = [
    'velocities', 'gears', 'times', 'on_engine', 'gear_box_type',
    'accelerations', 'tyre_code', 'after_treatment_speeds_delta',
    'motive_powers', 'accelerations', 'hybrid_modes', 'motor_p4_front_powers',
    'motor_p4_rear_powers', 'motor_p3_front_powers', 'motor_p3_rear_powers',
]

#: Relevant outputs of the model.
outputs = ['gear_box_speeds_in']

#: Targets to compare the outputs of the model.
targets = ['engine_speeds_out']

#: Extra inputs for the metrics.
metrics_inputs = [
    'times', 'velocities', 'gear_shifts', 'on_engine', 'stop_velocity',
    'hybrid_modes', 'after_treatment_warm_up_phases'
]


def metric_engine_speed_model(
        y_true, y_pred, times, velocities, gear_shifts, on_engine,
        stop_velocity, hybrid_modes, after_treatment_warm_up_phases=False):
    """
    Metric for the `engine_speed_model`.

    :param y_true:
        Reference engine speed vector [RPM].
    :type y_true: numpy.array

    :param y_pred:
        Predicted engine speed vector [RPM].
    :type y_pred: numpy.array

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param gear_shifts:
        When there is a gear shifting [-].
    :type gear_shifts: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :param hybrid_modes:
        Hybrid mode status (0: EV, 1: Parallel, 2: Serial).
    :type hybrid_modes: numpy.array

    :param after_treatment_warm_up_phases:
        Phases when engine speed is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :return:
        Error.
    :rtype: float
    """
    from co2mpas.utils import mae
    b = ~calculate_clutch_phases(times, 1, 1, gear_shifts, 0, (-4.0, 4.0))
    b &= (velocities > stop_velocity) & ~after_treatment_warm_up_phases
    b &= on_engine & (hybrid_modes == 1)
    return b.any() and float(mae(y_true[b], y_pred[b])) or .0


#: Metrics to compare outputs with targets.
metrics = sh.map_list(targets, metric_engine_speed_model)

#: Upper score limits to raise the warnings.
up_limit = sh.map_list(targets, 40)


def select_models(keys, data):
    """
    Select models from data.

    :param keys:
        Model keys.
    :type keys: list

    :param data:
        Cycle data.
    :type data: dict

    :return:
        Models.
    :rtype: dict
    """
    mdl = sh.selector(keys, data, allow_miss=True)
    if 'tyre_dynamic_rolling_coefficient' in mdl:
        mdl.pop('r_dynamic', None)
    return mdl


#: Prediction model.
# noinspection PyProtectedMember
dsp = sh.Blueprint(_physical, inputs, outputs, models)._set_cls(
    define_sub_model
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the mechanic of the wheels.
"""
import math
import regex
import schedula as sh
from co2mpas.defaults import dfl

dsp = sh.BlueDispatcher(
    name='Wheel model', description='It models the wheel dynamics.'
)


@sh.add_function(dsp, outputs=['tyre_code'], weight=11)
def define_tyre_code_v1(wheel_drive, front_tyre_code):
    """
    Defines the tyre code (e.g.,P225/70R14).

    :param wheel_drive:
        Wheel drive (i.e., front, rear, front+rear).
    :type wheel_drive: str

    :param front_tyre_code:
        Front wheel tyre code (e.g.,P225/70R14).
    :type front_tyre_code: str

    :return:
        Tyre code (e.g.,P225/70R14).
    :rtype: str
    """
    if 'front' in wheel_drive:
        return front_tyre_code
    return sh.NONE


@sh.add_function(dsp, outputs=['tyre_code'], weight=10)
def define_tyre_code(wheel_drive, rear_tyre_code):
    """
    Defines the tyre code (e.g.,P225/70R14).

    :param wheel_drive:
        Wheel drive (i.e., front, rear, front+rear).
    :type wheel_drive: str

    :param rear_tyre_code:
        Rear wheel tyre code (e.g.,P225/70R14).
    :type rear_tyre_code: str

    :return:
        Tyre code (e.g.,P225/70R14).
    :rtype: str
    """
    if 'rear' in wheel_drive:
        return rear_tyre_code
    return sh.NONE


def calculate_wheel_power(velocities, accelerations, road_loads, vehicle_mass):
    """
    Calculates the wheel power [kW].

    :param velocities:
        Velocity [km/h].
    :type velocities: numpy.array | float

    :param accelerations:
        Acceleration [m/s2].
    :type accelerations: numpy.array | float

    :param road_loads:
        Cycle road loads [N, N/(km/h), N/(km/h)^2].
    :type road_loads: list, tuple

    :param vehicle_mass:
        Vehicle mass [kg].
    :type vehicle_mass: float

    :return:
        Power at wheels [kW].
    :rtype: numpy.array | float
    """

    f0, f1, f2 = road_loads

    quadratic_term = f0 + (f1 + f2 * velocities) * velocities

    vel = velocities / 3600

    return (quadratic_term + 1.03 * vehicle_mass * accelerations) * vel


# noinspection PyIncorrectDocstring,SpellCheckingInspection
@sh.add_function(dsp, outputs=['wheel_torques'])
def calculate_wheel_torques(wheel_powers, wheel_speeds, coef=30000 / math.pi):
    """
    Calculates torque at the wheels [N*m].

    :param wheel_powers:
        Power at the wheels [kW].
    :type wheel_powers: numpy.array | float

    :param wheel_speeds:
        Rotating speed of the wheel [RPM].
    :type wheel_speeds: numpy.array | float

    :return:
        Torque at the wheels [N*m].
    :rtype: numpy.array | float
    """
    import numpy as np
    with np.errstate(divide='ignore', invalid='ignore'):
        torques = np.divide(wheel_powers, wheel_speeds) * coef
        return np.where(wheel_speeds, torques, 0)


@sh.add_function(dsp, outputs=['wheel_powers'])
def calculate_wheel_powers(wheel_torques, wheel_speeds):
    """
    Calculates power at the wheels [kW].

    :param wheel_torques:
        Torque at the wheel [N*m].
    :type wheel_torques: numpy.array | float

    :param wheel_speeds:
        Rotating speed of the wheel [RPM].
    :type wheel_speeds: numpy.array | float

    :return:
        Power at the wheels [kW].
    :rtype: numpy.array | float
    """

    return wheel_torques * wheel_speeds * (math.pi / 30000.0)


def _compile_speed_function(r_dynamic):
    c = 30.0 / (3.6 * math.pi * r_dynamic)

    def _wheel_speed(velocities):
        return velocities * c

    return _wheel_speed


@sh.add_function(dsp, outputs=['wheel_speeds'])
def calculate_wheel_speeds(velocities, r_dynamic):
    """
    Calculates rotating speed of the wheels [RPM].

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array | float

    :param r_dynamic:
        Dynamic radius of the wheels [m].
    :type r_dynamic: float

    :return:
        Rotating speed of the wheel [RPM].
    :rtype: numpy.array | float
    """
    return _compile_speed_function(r_dynamic)(velocities)


@sh.add_function(dsp, outputs=['r_dynamic'])
def identify_r_dynamic(
        velocity_speed_ratios, gear_box_ratios, final_drive_ratios):
    """
    Identifies the dynamic radius of the wheels [m].

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param gear_box_ratios:
        Gear box ratios [-].
    :type gear_box_ratios: dict[int, float | int]

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :return:
        Dynamic radius of the wheels [m].
    :rtype: float
    """
    from co2mpas.utils import reject_outliers
    from .gear_box.mechanical import calculate_speed_velocity_ratios
    svr = calculate_speed_velocity_ratios(
        gear_box_ratios, final_drive_ratios, 1
    )

    r = [svr[k] * vs for k, vs in velocity_speed_ratios.items() if k]

    r_dynamic = reject_outliers(r)[0]

    return r_dynamic


dsp.add_data('stop_velocity', dfl.values.stop_velocity)


@sh.add_function(dsp, outputs=['r_dynamic'], weight=10)
def identify_r_dynamic_v1(
        velocities, gears, gear_box_speeds_in, gear_box_ratios,
        final_drive_ratios, stop_velocity):
    """
    Identifies the dynamic radius of the wheels [m].

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array

    :param gear_box_ratios:
        Gear box ratios [-].
    :type gear_box_ratios: dict[int, float | int]

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Dynamic radius of the wheels [m].
    :rtype: float
    """
    import numpy as np
    from co2mpas.utils import reject_outliers
    from .gear_box import mechanical as gb_mec
    svr = gb_mec.calculate_speed_velocity_ratios(
        gear_box_ratios, final_drive_ratios, 1.0
    )

    vsr = gb_mec.calculate_velocity_speed_ratios(svr)

    speed_x_r_dyn_ratios = gb_mec.calculate_gear_box_speeds_in(
        gears, velocities, vsr, stop_velocity
    )
    with np.errstate(divide='ignore', invalid='ignore'):
        r_dynamic = speed_x_r_dyn_ratios / gear_box_speeds_in
    r_dynamic = r_dynamic[np.isfinite(r_dynamic)]
    r_dynamic = reject_outliers(r_dynamic)[0]
    return r_dynamic


@sh.add_function(dsp, outputs=['r_dynamic'], weight=11)
def identify_r_dynamic_v2(
        velocities, gears, engine_speeds_out, gear_box_ratios,
        final_drive_ratios, stop_velocity, on_engine):
    """
    Identifies the dynamic radius of the wheels [m].

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param gear_box_ratios:
        Gear box ratios [-].
    :type gear_box_ratios: dict[int, float | int]

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :return:
        Dynamic radius of the wheels [m].
    :rtype: float
    """
    return identify_r_dynamic_v1(
        velocities[on_engine], gears[on_engine], engine_speeds_out[on_engine],
        gear_box_ratios, final_drive_ratios, stop_velocity
    )


dsp.add_data('plateau_acceleration', dfl.values.plateau_acceleration)
dsp.add_data('change_gear_window_width', dfl.values.change_gear_window_width)


@sh.add_function(dsp, outputs=['r_dynamic'], weight=12)
def identify_r_dynamic_v3(
        times, velocities, accelerations, r_wheels, gear_box_speeds_in,
        gear_box_ratios, final_drive_ratios, idle_engine_speed, stop_velocity,
        plateau_acceleration, change_gear_window_width):
    """
    Identifies the dynamic radius of the wheels [m].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Vehicle acceleration [m/s2].
    :type accelerations: numpy.array

    :param r_wheels:
        Radius of the wheels [m].
    :type r_wheels: float

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array

    :param gear_box_ratios:
        Gear box ratios [-].
    :type gear_box_ratios: dict[int, float | int]

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :param plateau_acceleration:
        Maximum acceleration to be at constant velocity [m/s2].
    :type plateau_acceleration: float

    :param change_gear_window_width:
        Time window used to apply gear change filters [s].
    :type change_gear_window_width: float

    :return:
        Dynamic radius of the wheels [m].
    :rtype: float
    """
    from .gear_box import mechanical as gb_mec
    svr = gb_mec.calculate_speed_velocity_ratios(
        gear_box_ratios, final_drive_ratios, r_wheels
    )

    gears = gb_mec.identify_gears(
        times, velocities, accelerations, gear_box_speeds_in,
        gb_mec.calculate_velocity_speed_ratios(svr), stop_velocity,
        plateau_acceleration, change_gear_window_width, idle_engine_speed
    )

    r_dynamic = identify_r_dynamic_v1(
        velocities, gears, gear_box_speeds_in, gear_box_ratios,
        final_drive_ratios, stop_velocity
    )

    return r_dynamic


@sh.add_function(dsp, outputs=['r_dynamic'], weight=13)
def identify_r_dynamic_v4(
        times, velocities, accelerations, r_wheels, engine_speeds_out,
        gear_box_ratios, final_drive_ratios, idle_engine_speed, stop_velocity,
        plateau_acceleration, change_gear_window_width, on_engine):
    """
    Identifies the dynamic radius of the wheels [m].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Vehicle acceleration [m/s2].
    :type accelerations: numpy.array

    :param r_wheels:
        Radius of the wheels [m].
    :type r_wheels: float

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param gear_box_ratios:
        Gear box ratios [-].
    :type gear_box_ratios: dict[int, float | int]

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :param plateau_acceleration:
        Maximum acceleration to be at constant velocity [m/s2].
    :type plateau_acceleration: float

    :param change_gear_window_width:
        Time window used to apply gear change filters [s].
    :type change_gear_window_width: float

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :return:
        Dynamic radius of the wheels [m].
    :rtype: float
    """
    return identify_r_dynamic_v3(
        times[on_engine], velocities[on_engine], accelerations[on_engine],
        r_wheels, engine_speeds_out[on_engine], gear_box_ratios,
        final_drive_ratios, idle_engine_speed, stop_velocity,
        plateau_acceleration, change_gear_window_width

    )


_re_tyre_code_iso = regex.compile(
    r"""
    ^(?P<use>([a-z]){1,2})?\s*
    (?P<nominal_section_width>(\d){3})\s*
    \/\s*
    (?P<aspect_ratio>(\d){2,3})?
    ((\s*(?P<speed_rating>[a-z])?\s*(?P<carcass>[a-z])\s*)|\s+)
    (?P<rim_diameter>(\d){1,2}(\.(\d){1,2})?)
    (\s+(?P<use>C))?
    (\s+(?P<load_index>(\d){2,3}(/(\d){2,3})?)\s*
     (?P<speed_rating>(\([a-z]\)|[a-z]\d?)))?\s*
    (\s*((?P<load_range>[a-z])(^| )))?
    (\s+(?P<additional_marks>.*))?$
    """, regex.IGNORECASE | regex.X | regex.DOTALL)

_re_tyre_code_numeric = regex.compile(
    r"""
    ^((?P<diameter>(\d){2})\s*x\s*)?
    (?P<nominal_section_width>(\d){1,2}(\.(\d){1,2})?)\s*
    ((\s*(?P<carcass>([a-z]|-))\s*)|\s+)
    (?P<rim_diameter>(\d){2}(\.(\d){1,2})?)\s*
    (?P<use>(LT|C))\s*
    ((?P<load_index>(\d){2,3}(/(\d){2,3})?)\s*
     (?P<speed_rating>(\([a-z]\)|[a-z]\d?)))?\s*
    (\s*((?P<load_range>[a-z])(^| )))?
    (\s+(?P<additional_marks>.*))?$
    """, regex.IGNORECASE | regex.X | regex.DOTALL)

_re_tyre_code_pax = regex.compile(
    r"""
    ^(?P<use>([a-z]){1,2})?\s*
    (?P<nominal_section_width>(\d){3})\s*-\s*
    (?P<diameter>(\d){2,3})
    ((\s*(?P<carcass>[a-z])\s*)|\s+)
    (?P<rim_diameter>(\d){2,3})
    ((\s*(?P<load_range>[a-z])?\s*)|\s+)
    (\s*(?P<load_index>(\d){2,3})\s*
     (?P<speed_rating>[a-z]))?\s*
    (\s*(?P<additional_marks>.*))?
    """, regex.IGNORECASE | regex.X | regex.DOTALL)

_re_tyre_code_euro = regex.compile(
    r"""
    ^(?P<use>([a-z]){1,2})?\s*
    (?P<nominal_section_width>(\d){3})\s*
    ((\s*(?P<speed_rating>[a-z])?\s*(?P<carcass>[a-z])\s*)|\s+)
    (?P<rim_diameter>(\d){1,2}(\.(\d){1,2})?)
    (\s+(?P<additional_marks>.*))?$
    """, regex.IGNORECASE | regex.X | regex.DOTALL)


def _format_tyre_dimensions(tyre_dimensions):
    import schema
    frt = schema.Schema({
        schema.Optional('additional_marks'): schema.Use(str),
        schema.Optional('aspect_ratio'): schema.Use(float),
        schema.Optional('carcass'): schema.Use(str),
        'rim_diameter': schema.Use(float),
        schema.Optional('diameter'): schema.Use(float),
        schema.Optional('load_index'): schema.Use(str),
        schema.Optional('load_range'): schema.Use(str),
        'nominal_section_width': schema.Use(float),
        schema.Optional('speed_rating'): schema.Use(str),
        schema.Optional('use'): schema.Use(str),
        schema.Optional('code'): schema.Use(str),
    })
    m = {k: v for k, v in tyre_dimensions.items() if v is not None}
    return frt.validate(m)


@sh.add_function(dsp, outputs=['tyre_dimensions'])
def calculate_tyre_dimensions(tyre_code):
    """
    Calculates the tyre dimensions from the tyre code.

    :param tyre_code:
        Tyre code (e.g.,P225/70R14).
    :type tyre_code: str

    :return:
        Tyre dimensions.
    :rtype: dict
    """
    import schema
    it = [
        ('iso', _re_tyre_code_iso),
        ('numeric', _re_tyre_code_numeric),
        ('pax', _re_tyre_code_pax),
        ('euro', _re_tyre_code_euro)
    ]
    for c, _r in it:
        try:
            m = _r.match(tyre_code).groupdict()
            m['code'] = c
            if c == 'numeric' and 'aspect_ratio' not in m:
                b = m['nominal_section_width'].split('.')[-1][-1] == '5'
                m['aspect_ratio'] = '82' if b else '92'
            elif c == 'euro' and 'aspect_ratio' not in m:
                m['aspect_ratio'] = '80'
            return _format_tyre_dimensions(m)
        except (AttributeError, schema.SchemaError):
            pass
    raise ValueError('Invalid tyre code: %s', tyre_code)


# noinspection PyUnusedLocal
def _format_tyre_code(
        nominal_section_width, rim_diameter, aspect_ratio=0, use='', carcass='',
        load_index='', speed_rating='', additional_marks='', load_range='',
        code='iso', diameter=None, **kw):
    if code == 'iso':
        parts = (
            '%s%d/%d%s%d' % (use, nominal_section_width, aspect_ratio,
                             carcass or ' ', rim_diameter),
        )
    elif code == 'numeric':
        diameter = '%.2fx' % diameter if diameter is not None else ''
        parts = (
            '%s%.2f%s%.2f %s' % (diameter, nominal_section_width,
                                 carcass or ' ', rim_diameter, use),
        )
    else:
        parts = (
            '%d-%d%s%d%s' % (nominal_section_width, diameter, carcass or ' ',
                             rim_diameter, load_range),
        )

    parts += (
        '%s%s' % (load_index, speed_rating),
        load_range,
        additional_marks
    )
    return ' '.join(p for p in parts if p)


@sh.add_function(dsp, outputs=['tyre_code'])
def define_tyre_code(tyre_dimensions):
    """
    Returns the tyre code from the tyre dimensions.

    :param tyre_dimensions:
        Tyre dimensions.

        .. note:: The fields are : use, nominal_section_width, aspect_ratio,
           carcass, diameter, load_index, speed_rating, and additional_marks.
    :type tyre_dimensions: dict

    :return:
        Tyre code (e.g.,P225/70R14).
    :rtype: str
    """
    return _format_tyre_code(**tyre_dimensions)


@sh.add_function(dsp, outputs=['tyre_code'], weight=5)
def default_tyre_code(r_dynamic):
    """
    Return one of the most popular tyre code according to the r dynamic.

    :param r_dynamic:
        Dynamic radius of the wheels [m].
    :type r_dynamic: float

    :return:
        Tyre code (e.g.,P225/70R14).
    :rtype: str
    """

    pop = (
        '165/65R13', '155/70R13', '165/70R13', '165/60R14', '185/60R14',
        '155/65R14', '165/65R14', '175/65R14', '185/65R14', '165/70R14',
        '175/70R14', '195/50R15', '185/55R15', '195/55R15', '185/60R15',
        '195/60R15', '205/60R15', '175/65R15', '185/65R15', '195/65R15',
        '195/70R15', '195/45R16', '205/45R16', '205/50R16', '195/55R16',
        '205/55R16', '215/55R16', '205/60R16', '215/60R16', '215/65R16',
        '205/40R17', '245/40R17', '205/45R17', '215/45R17', '225/45R17',
        '235/45R17', '205/50R17', '215/50R17', '225/50R17', '215/55R17',
        '225/55R17', '215/60R17', '225/65R17', '235/65R17', '225/40R18',
        '235/40R18', '245/40R18', '225/45R18', '235/60R18', '255/35R19'
    )

    def _key_func(c):
        r = calculate_r_wheels(calculate_tyre_dimensions(c))
        return r <= r_dynamic, (r - r_dynamic) ** 2

    return min(pop, key=_key_func)


@sh.add_function(dsp, outputs=['r_wheels'])
def calculate_r_wheels(tyre_dimensions):
    """
    Calculates the radius of the wheels [m] from the tyre dimensions.

    :param tyre_dimensions:
        Tyre dimensions.

        .. note:: The fields are : use, nominal_section_width, aspect_ratio,
           carcass, diameter, load_index, speed_rating, and additional_marks.
    :type tyre_dimensions: dict

    :return:
        Radius of the wheels [m].
    :rtype: float
    """
    if 'diameter' in tyre_dimensions:
        if tyre_dimensions['code'] == 'pax':
            return tyre_dimensions['diameter'] / 2000  # Diameter is in mm.
        return tyre_dimensions['diameter'] * 0.0254  # Diameter is in inches.
    a = tyre_dimensions['aspect_ratio'] / 100  # Aspect ratio is Height/Width.
    w = tyre_dimensions['nominal_section_width']
    if tyre_dimensions.get('code', 'iso') == 'iso':
        w /= 1000  # Width is in mm.
    else:
        w *= 0.0254  # Width is in inches.

    dr = tyre_dimensions['rim_diameter'] * 0.0254  # Rim is in inches.
    return a * w + dr / 2


dsp.add_data(
    data_id='tyre_dynamic_rolling_coefficient', initial_dist=50,
    default_value=dfl.values.tyre_dynamic_rolling_coefficient,
)


@sh.add_function(dsp, outputs=['r_dynamic'])
def calculate_r_dynamic(r_wheels, tyre_dynamic_rolling_coefficient):
    """
    Calculates the dynamic radius of the wheels [m].

    :param r_wheels:
        Radius of the wheels [m].
    :type r_wheels: float

    :param tyre_dynamic_rolling_coefficient:
        Dynamic rolling coefficient [-].
    :type tyre_dynamic_rolling_coefficient: float

    :return:
        Dynamic radius of the wheels [m].
    :rtype: float
    """

    return tyre_dynamic_rolling_coefficient * r_wheels


@sh.add_function(dsp, outputs=['tyre_dynamic_rolling_coefficient'])
def identify_tyre_dynamic_rolling_coefficient(r_wheels, r_dynamic):
    """
    Identifies the dynamic rolling coefficient [-].

    :param r_wheels:
        Radius of the wheels [m].
    :type r_wheels: float

    :param r_dynamic:
        Dynamic radius of the wheels [m].
    :type r_dynamic: float

    :return:
        Dynamic rolling coefficient [-].
    :rtype: float
    """

    return r_dynamic / r_wheels


dsp.add_function(
    function=sh.bypass, inputs=['motive_powers'], outputs=['wheel_powers']
)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to define theoretical profiles of NEDC cycle.
"""
import schedula as sh
from co2mpas.defaults import dfl

dsp = sh.BlueDispatcher(
    name='NEDC cycle model',
    description='Returns the theoretical times, velocities, and gears of NEDC.'
)

dsp.add_data(
    'initial_temperature', dfl.values.initial_temperature_NEDC,
    description='Initial temperature of the test cell [°C].'
)
dsp.add_data(
    'max_time', dfl.values.max_time_NEDC, 5, description='Maximum time [s].',

)
dsp.add_data('k1', dfl.values.k1)
dsp.add_data('k2', dfl.values.k2)
dsp.add_data('k5', dfl.values.k5, 10)

dsp.add_function(
    function_id='set_max_gear_as_default_k5',
    function=sh.bypass,
    inputs=['max_gear'],
    outputs=['k5']
)


def nedc_gears(times, max_gear, k1=1, k2=2, k5=2):
    """
    Returns the gear shifting profile according to NEDC [-].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param max_gear:
        Maximum gear of the gear box [-].
    :type max_gear: int

    :param k1:
        K1 NEDC parameter (first or second gear) [-].
    :type k1: int

    :param k2:
        K2 NEDC parameter (first or second gear) [-].
    :type k2: int

    :param k5:
        K5 NEDC parameter (first or second gear) [-].
    :type k5: int

    :return:
        Gear vector [-].
    :rtype: numpy.array
    """
    from scipy.interpolate import interp1d
    eps, parts = 0.01, {}
    # part one
    parts[0] = parts[1] = parts[2] = parts[3] = (
        (0, 0), (6, 0), (6, k1), (11, k1), (11, 1), (25, 1), (25, k1), (28, k1),
        (28, 0), (44, 0), (44, k1), (49, k1), (49, 1), (55 - eps, 1),
        (55 - eps, 2), (93, 2), (93, k2), (96, k2), (96, 0), (112, 0),
        (112, k1), (117, k1), (117, 1), (123 - eps, 1), (123 - eps, 2),
        (134 - eps, 2), (134 - eps, 3), (177 - eps, 3), (177 - eps, 2),
        (185, 2), (185, k2), (188, k2), (188, 0), (195, 0)
    )

    # part two
    parts[4] = (
        (0, k1), (20, k1), (20, 1), (26 - eps, 1), (26 - eps, 2), (37 - eps, 2),
        (37 - eps, 3), (47 - eps, 3), (47 - eps, 4), (61, 4), (61, 5), (115, 5),
        (115, 4), (201, 4), (201, 5), (286, 5), (286, max_gear),
        (370, max_gear), (370, k5), (380, k5), (380, 0), (400, 0)
    )

    tg = _repeat_parts(times, parts).T
    s = interp1d(*tg, kind='nearest', assume_sorted=True)(times)
    s[s > max_gear] = max_gear
    return s


# noinspection PyUnusedLocal, PyMissingOrEmptyDocstring
def is_manual(gear_box_type, *args):
    return gear_box_type == 'manual'


dsp.add_function(
    function_id='nedc_gears',
    function=sh.add_args(nedc_gears),
    inputs=['gear_box_type', 'times', 'max_gear', 'k1', 'k2', 'k5'],
    outputs=['gears'],
    input_domain=is_manual
)


def _repeat_parts(times, parts):
    import numpy as np
    from functools import reduce
    it = [v[1] for v in sorted(parts.items())]

    def _func(x, y):
        y = np.array(y)
        y[:, 0] += x[-1][-1][0]
        return x + (y,)

    tv = np.concatenate(reduce(_func, it[1:], (it[0],)))
    n = int(np.ceil(times[-1] / tv[-1][0])) - 1
    if n > 0:
        tv = np.concatenate(reduce(_func, (tv,) * n, (tv,)))

    return tv


@sh.add_function(dsp, outputs=['velocities'])
def nedc_velocities(times, gear_box_type):
    """
    Returns the velocity profile according to NEDC and gear box type [km/h].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param gear_box_type:
        Gear box type (manual or automatic or cvt).
    :type gear_box_type: str

    :return:
        Velocity vector [km/h].
    :rtype: numpy.array
    """
    from numpy import interp
    parts = {
        'manual':
            {
                0: (
                    (0, 0), (11, 0), (15, 15), (23, 15), (25, 10), (28, 0),
                    (49, 0), (54, 15), (56, 15), (61, 32), (85, 32), (93, 10),
                    (96, 0), (117, 0), (122, 15), (124, 15), (133, 35),
                    (135, 35), (143, 50), (155, 50), (163, 35), (176, 35),
                    (178, 35), (185, 10), (188, 0), (195, 0)
                ),
                4: (
                    (0, 0), (20, 0), (25, 15), (27, 15), (36, 35), (38, 35),
                    (46, 50), (48, 50), (61, 70), (111, 70), (119, 50),
                    (188, 50), (201, 70), (251, 70), (286, 100), (316, 100),
                    (336, 120), (346, 120), (362, 80), (370, 50), (380, 0),
                    (400, 0)
                )
            },
        'automatic':
            {
                0: (
                    (0, 0), (11, 0), (15, 15), (23, 15), (25, 10), (28, 0),
                    (49, 0), (61, 32), (85, 32), (93, 10), (96, 0), (117, 0),
                    (143, 50), (155, 50), (163, 35), (176, 35), (178, 35),
                    (185, 10), (188, 0), (195, 0)
                ),
                4: (
                    (0, 0), (20, 0), (61, 70), (111, 70), (119, 50), (188, 50),
                    (201, 70), (251, 70), (286, 100), (316, 100), (336, 120),
                    (346, 120), (362, 80), (370, 50), (380, 0), (400, 0)
                )
            }
    }

    for k, v in parts.items():
        v[1] = v[2] = v[3] = v[0]

    parts['cvt'] = parts['planetary'] = parts['none'] = parts['automatic']

    tv = _repeat_parts(times, parts[gear_box_type])

    return interp(times, *tv.T)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the manual gear shifting.
"""
import collections
import numpy as np
import schedula as sh
from .at_gear import CMV

dsp = sh.BlueDispatcher(
    name='manual GS model', description='Models the manual gear shifting.'
)


@sh.add_function(dsp, outputs=['engine_max_speed_95'])
def calculate_engine_max_speed_95(
        full_load_speeds, idle_engine_speed, engine_max_speed, full_load_curve,
        engine_max_power):
    """
    Calculates the maximum engine speed [RPM] at 95% of the nominal power.

    :param full_load_speeds:
        T1 map speed vector [RPM].
    :type full_load_speeds: numpy.array

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param engine_max_speed:
        Maximum allowed engine speed [RPM].
    :type engine_max_speed: float

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :param engine_max_power:
        Engine nominal power [kW].
    :type engine_max_power: float

    :return:
        Maximum engine speed [RPM] at 95% of the nominal power.
    :rtype: float
    """
    from scipy.interpolate import InterpolatedUnivariateSpline as Spl
    speeds = [idle_engine_speed[0] - idle_engine_speed[1], engine_max_speed]
    speeds.extend(full_load_speeds)
    speeds = np.unique(speeds)
    n = [engine_max_speed]
    n.extend(
        Spl(speeds, full_load_curve(speeds) / engine_max_power - 0.95).roots()
    )

    return max(n)


# noinspection PyMissingOrEmptyDocstring
class MGS(CMV):
    def __init__(self, *args, **kwargs):
        super(MGS, self).__init__(*args, **kwargs)

    # noinspection PyMethodOverriding,PyMethodOverriding
    def fit(self, full_load_curve, engine_speed_at_max_power, road_loads,
            engine_max_speed_95, velocity_speed_ratios, idle_engine_speed):
        self.velocity_speed_ratios = velocity_speed_ratios
        # noinspection PyProtectedMember
        from co2mpas.defaults import dfl
        from .mechanical import _calculate_req_power as calc_power_req

        d = dfl.functions.MGS
        vel = np.arange(d.MIN_VEL, d.MAX_VEL, d.DELTA_VEL, float)[:, None]

        g_id, vsr = zip(*[(k, v) for k, v in sorted(
            velocity_speed_ratios.items(), reverse=True
        ) if k])
        s = np.round(np.divide(vel, vsr), 1)

        p = full_load_curve(s)

        b = np.zeros_like(p, bool)
        g_id, idle = np.array(g_id), idle_engine_speed[0]
        j = g_id == 1
        b[:, j] = s[:, j] < idle
        j = g_id == 2
        b[:, j] = s[:, j] < idle * 0.9
        j = np.in1d(g_id, (1, 2), True, True)
        b[:, j] = s[:, j] < idle + 0.125 * (engine_speed_at_max_power - idle)

        b |= s > engine_max_speed_95
        b[:, j] |= p[:, j] < (calc_power_req(road_loads, vel) / d.PREC_FLC)

        v = np.repeat(vel, b.shape[1], 1)
        v[b] = np.nan
        d, u = np.nanmin(v, 0), np.nanmax(v, 0)
        d[-1], u[0] = 0, dfl.INF

        self.clear()
        self.update(collections.OrderedDict(zip(g_id, zip(d, u))))
        self[0] = (0, 0)
        return self


@sh.add_function(dsp, outputs=['MGS', 'MVL'])
def define_msg_and_mvl(
        full_load_curve, road_loads, velocity_speed_ratios,
        engine_max_speed_95, idle_engine_speed, engine_speed_at_max_power):
    """
    Calculates the maximum velocity from full load curve.

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :param road_loads:
        Cycle road loads [N, N/(km/h), N/(km/h)^2].
    :type road_loads: list, tuple

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]
    
    :param engine_max_speed_95:
        Maximum engine speed [RPM] at 95% of the nominal power.
    :type engine_max_speed_95: float
    
    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)
    
    :param engine_speed_at_max_power:
        Engine speed at engine nominal power [RPM].
    :type engine_speed_at_max_power: float

    :return:
        Maximum velocity and gear at maximum velocity [km/h, -].
    :return: float, int
    """
    from .at_gear import MVL
    mgs = MGS().fit(
        full_load_curve, engine_speed_at_max_power, road_loads,
        engine_max_speed_95, velocity_speed_ratios, idle_engine_speed
    )
    mvl = MVL()
    mvl.update(mgs)
    mvl.velocity_speed_ratios = velocity_speed_ratios

    return mgs, mvl


@sh.add_function(
    dsp,
    inputs=[
        'cycle_type', 'velocity_speed_ratios', 'MVL', 'idle_engine_speed',
        'full_load_curve', 'engine_speed_at_max_power'
    ],
    outputs=['correct_gear']
)
def correct_gear_v4(
        cycle_type, velocity_speed_ratios, mvl, idle_engine_speed,
        full_load_curve, engine_speed_at_max_power):
    """
    Returns a function to correct the gear predicted according to
    :func:`correct_gear_mvl`, :func:`correct_gear_full_load` and driveability
    rules.

    :param cycle_type:
        Cycle type (WLTP or NEDC).
    :type cycle_type: str

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param mvl:
        Matrix velocity limits (upper and lower bound) [km/h].
    :type mvl: MVL

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :param engine_speed_at_max_power:
        Rated engine speed [RPM].
    :type engine_speed_at_max_power: float

    :return:
        A function to correct the predicted gear.
    :rtype: callable
    """
    from .at_gear import correct_gear_v0
    cg = correct_gear_v0(
        cycle_type, velocity_speed_ratios, mvl, idle_engine_speed,
        full_load_curve
    )
    cg.fit_correct_driveability_rules(engine_speed_at_max_power)
    return cg


@sh.add_function(dsp, outputs=['gears'], weight=sh.inf(1, 0), inputs=[
    'correct_gear', 'MGS', 'times', 'velocities', 'accelerations',
    'motive_powers'
])
def prediction_gears_gsm_v1(
        correct_gear, gsm, times, velocities, accelerations, motive_powers):
    """
    Predicts gears with a gear shifting model (mgs) [-].

    :param correct_gear:
        A function to correct the gear predicted.
    :type correct_gear: callable

    :param gsm:
        A gear shifting model (cmv or gspv or dtgs).
    :type gsm: MSG

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Vehicle acceleration [m/s2].
    :type accelerations: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :return:
        Predicted gears.
    :rtype: numpy.array
    """
    gears = gsm.predict(
        times, velocities, accelerations, motive_powers,
        correct_gear=correct_gear
    )
    return gears



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
CO2MPAS setup.
"""
import io
import os
import collections
import os.path as osp

name = 'co2mpas'
mydir = osp.dirname(__file__)


def read_project_version():
    """
    Version-trick to have version-info in a single place,
    taken from: http://stackoverflow.com/questions/2058802/how-can-i-get-the-
    version-defined-in-setup-py-setuptools-in-my-package

    :return:
        Project version.
    :rtype: str
    """
    fglobals = {}
    with io.open(osp.join(mydir, name, '_version.py'), encoding='UTF-8') as fd:
        exec(fd.read(), fglobals)  # To read __version__
    return fglobals['__version__']


# noinspection PyPackageRequirements
def get_long_description(cleanup=True):
    """
    Return the project long description.

    :param cleanup:
        Clean up the build folder.
    :type cleanup: bool

    :return:
        Project long description.
    :rtype: str
    """
    import shutil
    import tempfile
    from doc.conf import extensions
    from sphinx.application import Sphinx
    from sphinx.util.osutil import abspath
    from sphinxcontrib.writers.rst import RstTranslator
    from sphinx.ext.graphviz import text_visit_graphviz
    RstTranslator.visit_dsp = text_visit_graphviz
    outdir = tempfile.mkdtemp(prefix='setup-', dir='.')
    exclude_patterns = os.listdir(mydir or '.')
    exclude_patterns.remove('pypi.rst')

    # noinspection PyTypeChecker
    app = Sphinx(
        abspath(mydir), osp.join(mydir, 'doc/'), outdir, outdir + '/.doctree',
        'rst', confoverrides={
            'exclude_patterns': exclude_patterns,
            'master_doc': 'pypi',
            'dispatchers_out_dir': abspath(outdir + '/_dispatchers'),
            'extensions': extensions + ['sphinxcontrib.restbuilder']
        }, status=None, warning=None)

    app.build(filenames=[osp.join(app.srcdir, 'pypi.rst')])

    with open(outdir + '/pypi.rst') as file:
        res = file.read()

    if cleanup:
        shutil.rmtree(outdir)
    return res


if __name__ == '__main__':
    import functools
    from setuptools import setup, find_packages

    proj_ver = read_project_version()
    url = 'https://github.com/JRCSTU/%s-ta' % name
    download_url = '%s/tarball/v%s' % (url, proj_ver)
    project_urls = collections.OrderedDict((
        ('Documentation', 'http://%s.readthedocs.io' % name),
        ('Issue tracker', '%s/issues' % url),
    ))

    long_description = ''
    if os.environ.get('ENABLE_SETUP_LONG_DESCRIPTION') == 'TRUE':
        try:
            long_description = get_long_description()
            print('LONG DESCRIPTION ENABLED!')
        except Exception as ex:
            print('LONG DESCRIPTION ERROR:\n %r', ex)

    extras = {
        'cli': ['click', 'click-log'],
        'sync': ['syncing>=1.0.4', 'pandas>=0.21.0'],
        'plot': ['flask', 'regex', 'graphviz', 'Pygments', 'lxml',
                 'beautifulsoup4', 'jinja2', 'docutils', 'plotly'],
        'io': ['pandas>=0.21.0', 'dill', 'regex', 'xlref', 'xlrd', 'asteval']
    }
    extras['dice'] = ['co2mpas_dice>=4.0.5'] + extras['io']
    # noinspection PyTypeChecker
    extras['all'] = list(functools.reduce(set.union, extras.values(), set()))
    extras['dev'] = extras['all'] + [
        'wheel', 'sphinx', 'gitchangelog', 'mako', 'sphinx_rtd_theme',
        'setuptools>=36.0.1', 'sphinxcontrib-restbuilder', 'nose', 'coveralls',
        'ddt', 'sphinx-click'
    ]

    setup(
        name=name,
        version=proj_ver,
        packages=find_packages(exclude=[
            'test', 'test.*',
            'doc', 'doc.*',
            'appveyor'
        ]),
        url=url,
        project_urls=project_urls,
        download_url=download_url,
        license='EUPL 1.1+',
        author='CO2MPAS-Team',
        author_email='JRC-CO2MPAS@ec.europa.eu',
        description='The Type-Approving vehicle simulator predicting NEDC CO2 '
                    'emissions from WLTP',
        long_description=long_description,
        keywords="""CO2 fuel-consumption WLTP NEDC vehicle automotive EU JRC IET 
        STU correlation back-translation policy monitoring M1 N1 simulator 
        engineering scientific
        """.split(),
        classifiers=[
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: Implementation :: CPython",
            "Development Status :: 4 - Beta",
            'Natural Language :: English',
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Intended Audience :: Manufacturing",
            'Environment :: Console',
            'License :: OSI Approved :: European Union Public Licence 1.1 '
            '(EUPL 1.1)',
            'Natural Language :: English',
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: Unix",
            "Operating System :: OS Independent",
            'Topic :: Scientific/Engineering',
            "Topic :: Scientific/Engineering :: Information Analysis",
        ],
        obsoletes=['co2mpas (< 4.0)'],
        python_requires='>=3.5',
        install_requires=[
            'PyYAML',
            'schedula>=0.3.2',
            'tqdm',
            'scikit-learn',
            'regex',
            'lmfit>=0.9.7',
            'numpy',
            'schema',
            'scipy',
            'wltp',
            'xgboost>=0.90',
            'statsmodels',
            'syncing>=1.0.4'
        ],
        entry_points={
            'console_scripts': [
                '%(p)s = %(p)s.cli:cli' % {'p': name},
            ],
        },
        extras_require=extras,
        tests_require=['nose>=1.0', 'ddt'],
        test_suite='nose.collector',
        package_data={
            'co2mpas': [
                'demos/*.xlsx',
                'templates/*.xlsx'
            ]
        },
        zip_safe=True,
        options={'bdist_wheel': {'universal': True}},
        platforms=['any'],
    )



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the GSPV Approach with Cold/Hot.
"""
import schedula as sh
from co2mpas.defaults import dfl
from .core import prediction_gears_gsm, GSMColdHot

dsp = sh.BlueDispatcher(
    name='Gear Shifting Power Velocity Approach with Cold/Hot'
)


@sh.add_function(
    dsp, inputs_kwargs=True, inputs_defaults=True, outputs=['GSPV_Cold_Hot']
)
def calibrate_gspv_cold_hot(
        times, gears, velocities, motive_powers, velocity_speed_ratios,
        time_cold_hot_transition=dfl.values.time_cold_hot_transition,
        stop_velocity=dfl.values.stop_velocity):
    """
    Identifies gear shifting power velocity matrices for cold and hot phases.

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param time_cold_hot_transition:
        Time at cold hot transition phase [s].
    :type time_cold_hot_transition: float

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Gear shifting power velocity matrices for cold and hot phases.
    :rtype: dict
    """
    from .gspv import GSPV
    model = GSMColdHot(time_cold_hot_transition=time_cold_hot_transition).fit(
        GSPV, times, gears, velocities, motive_powers, velocity_speed_ratios,
        stop_velocity
    )

    return model or sh.NONE


# predict gears with corrected matrix velocity
dsp.add_function(
    function=prediction_gears_gsm,
    inputs=[
        'correct_gear', 'gear_filter', 'GSPV_Cold_Hot', 'times', 'velocities',
        'accelerations', 'motive_powers', 'cycle_type', 'velocity_speed_ratios'
    ],
    outputs=['gears']
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and constants to define the at_model selector.
"""
import logging
import collections
import schedula as sh
import co2mpas.utils as co2_utl
from ._core import define_sub_model, _accuracy_score
from ...physical.gear_box.at_gear import dsp as _at_gear
from ...physical.gear_box.mechanical import calculate_gear_box_speeds_in

log = logging.getLogger(__name__)

#: Model name.
name = 'at_model'

#: Parameters that constitute the model.
models = [
    'MVL', 'CMV', 'CMV_Cold_Hot', 'DTGS', 'GSPV', 'GSPV_Cold_Hot',
    'specific_gear_shifting', 'change_gear_window_width',
    'max_velocity_full_load_correction', 'plateau_acceleration'
]

#: Inputs required to run the model.
inputs = [
    'idle_engine_speed', 'full_load_curve', 'accelerations', 'motive_powers',
    'engine_speeds_out', 'engine_temperatures', 'plateau_acceleration',
    'time_cold_hot_transition', 'times', 'stop_velocity', 'cycle_type',
    'use_dt_gear_shifting', 'specific_gear_shifting', 'velocity_speed_ratios',
    'velocities', 'MVL', 'fuel_saving_at_strategy', 'change_gear_window_width',
    'max_velocity_full_load_correction'
]

#: Relevant outputs of the model.
outputs = ['gears']

#: Targets to compare the outputs of the model.
targets = outputs

#: Weights coefficients to compute the model score.
weights = sh.map_list(targets, -1)

#: Metrics to compare outputs with targets.
metrics = {'gears': _accuracy_score}


def _correlation_coefficient(t, o):
    import numpy as np
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.corrcoef(t, o)[0, 1] if t.size > 1 else np.nan


def calculate_error_coefficients(
        identified_gears, gears, engine_speeds, predicted_engine_speeds,
        velocities, stop_velocity):
    """
    Calculates the prediction's error coefficients.

    :param identified_gears:
        Identified gear vector [-].
    :type identified_gears: numpy.array

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param engine_speeds:
        Engine speed vector [RPM].
    :type engine_speeds: numpy.array

    :param predicted_engine_speeds:
        Predicted Engine speed vector [RPM].
    :type predicted_engine_speeds: numpy.array

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Correlation coefficient and mean absolute error.
    :rtype: dict
    """

    b = velocities > stop_velocity

    x = engine_speeds[b]
    y = predicted_engine_speeds[b]
    res = {
        'mean_absolute_error': co2_utl.mae(x, y),
        'correlation_coefficient': _correlation_coefficient(x, y),
        'accuracy_score': _accuracy_score(identified_gears, gears)
    }

    return res


def select_models(keys, data):
    """
    Select models from data.

    :param keys:
        Model keys.
    :type keys: list

    :param data:
        Cycle data.
    :type data: dict

    :return:
        Models.
    :rtype: dict
    """
    sgs = 'specific_gear_shifting'
    # Namespace shortcuts.
    try:
        vel, vsr = data['velocities'], data['velocity_speed_ratios']
        t_eng, t_gears = data['engine_speeds_out'], data['gears']
        sv, at_m = data['stop_velocity'], data[sgs]
    except KeyError:
        return {}

    t_e = 'mean_absolute_error', 'accuracy_score', 'correlation_coefficient'

    # at_models to be assessed.
    if at_m == 'ALL':
        at_m = {'CMV', 'CMV_Cold_Hot', 'DTGS', 'GSPV', 'GSPV_Cold_Hot'}
    else:
        at_m = {at_m}

    # Other models to be taken from calibration output.
    mdl = sh.selector(set(keys) - at_m, data, allow_miss=True)

    # Inputs to predict the gears.
    inp = sh.selector(inputs, data, allow_miss=True)

    func = _at_gear.register()

    def _err(model_id, model):
        gears = func(
            inputs=sh.combine_dicts(inp, {sgs: model_id, model_id: model}),
            outputs=['gears']
        )['gears']

        eng = calculate_gear_box_speeds_in(gears, vel, vsr, sv)
        return calculate_error_coefficients(t_gears, gears, t_eng, eng, vel, sv)

    def _sort(v):
        e = sh.selector(t_e, v[0], output_type='list')
        return (e[0], -e[1], -e[2]), v[1]

    # Sort by error.
    rank = sorted((
        (_err(k, m), k, m)
        for k, m in sorted(sh.selector(at_m, data, allow_miss=True).items())
    ), key=_sort)

    if rank:
        data['at_scores'] = collections.OrderedDict(
            (k, e) for e, k, m in rank)
        e, k, m = rank[0]
        mdl[sgs], mdl[k] = k, m
        log.debug(
            'at_gear_shifting_model: %s with mean_absolute_error %.3f '
            '[RPM], accuracy_score %.3f, and correlation_coefficient '
            '%.3f.', k, *sh.selector(t_e, e, output_type='list'))

    return mdl


#: Prediction model.
# noinspection PyProtectedMember
dsp = sh.Blueprint(_at_gear, inputs, outputs, models)._set_cls(
    define_sub_model
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
It provides CO2MPAS schema parse/validator.
"""

import re
import pprint
import logging
import functools
import numpy as np
import os.path as osp
import schedula as sh
import co2mpas.utils as co2_utl
from collections import OrderedDict
from collections.abc import Iterable
from schema import Schema, Use, And, Or, SchemaError

log = logging.getLogger(__name__)

_re_1 = re.compile('(?<!})}(?!})')
_re_2 = re.compile('(?<!{){(?!{)')


def _format_error(error):
    if error:
        error = _re_1.sub('}}', error)
        error = _re_2.sub('{{', error)
    return error


# noinspection PyMissingOrEmptyDocstring
class Empty:
    def __repr__(self):
        return '%s' % self.__class__.__name__

    @staticmethod
    def validate(data):
        if isinstance(data, str) and data == 'EMPTY':
            return sh.EMPTY

        try:
            empty = not (data or data == 0)
        except ValueError:
            empty = np.isnan(data).all()

        if empty:
            return sh.NONE
        else:
            raise SchemaError('%r is not empty' % data)


# noinspection PyUnusedLocal
def _function(error=None, read=True, **kwargs):
    def _check_function(f):
        assert callable(f)
        return f

    if read:
        error = _format_error(error or 'should be a function!')
        return _eval(Use(_check_function), error=error)
    return And(_check_function, Use(lambda x: sh.NONE), error=error)


# noinspection PyUnusedLocal
def _string(error=None, **kwargs):
    error = _format_error(error or 'should be a string!')
    return Use(str, error=error)


# noinspection PyUnusedLocal
def _select(types=(), error=None, **kwargs):
    error = _format_error(error or 'should be one of {}!'.format(types))
    types = {k.lower(): k for k in types}
    return And(str, Use(lambda x: types[x.lower()]), error=error)


def _check_positive(x):
    return x >= 0


# noinspection PyUnusedLocal,PyShadowingBuiltins
def _positive(type=float, error=None, check=_check_positive, **kwargs):
    error = _format_error(error or 'should be as {} and positive!'.format(type))
    return And(Use(type), check, error=error)


# noinspection PyUnusedLocal
def _limits(limits=(0, 100), error=None, **kwargs):
    error = _format_error(error or 'should be {} <= x <= {}!'.format(*limits))

    def _check_limits(x):
        return limits[0] <= x <= limits[1]

    return And(Use(float), _check_limits, error=error)


_usersyms = {
    'no_model': lambda *x: 0
}


# noinspection PyUnusedLocal
def _eval(s, error=None, usersyms=None, **kwargs):
    error = _format_error(error or 'cannot be eval!')
    from asteval import Interpreter
    usersyms = sh.combine_dicts(_usersyms, usersyms or {})
    return Or(And(str, Use(Interpreter(usersyms=usersyms).eval), s), s,
              error=error)


# noinspection PyUnusedLocal,PyShadowingBuiltins
def _dict(format=None, error=None, read=True, pformat=pprint.pformat, **kwargs):
    format = And(dict, format or {int: float})
    error = error or 'should be a dict with this format {}!'.format(format)
    error = _format_error(error)
    c = Use(lambda x: {k: v for k, v in dict(x).items() if v is not None})
    if read:
        return _eval(Or(Empty(), And(c, Or(Empty(), format))), error=error)
    else:
        return And(_dict(format=format, error=error), Use(pformat))


# noinspection PyUnusedLocal,PyShadowingBuiltins
def _ordict(format=None, error=None, read=True, **kwargs):
    format = format or {int: float}
    msg = 'should be a OrderedDict with this format {}!'
    error = _format_error(error or msg.format(format))
    c = Use(OrderedDict)
    if read:
        return _eval(
            Or(Empty(), And(c, Or(Empty(), format))), error=error,
            usersyms={'OrderedDict': OrderedDict}
        )
    else:
        return And(_dict(format=format, error=error), Use(pprint.pformat))


def _check_length(length):
    if not isinstance(length, Iterable):
        length = (length,)

    def _check(data):
        ld = len(data)
        return any(ld == l for l in length)

    return _check


# noinspection PyUnusedLocal,PyShadowingBuiltins
def _type(type=None, error=None, length=None, **kwargs):
    type = type or tuple
    usersyms = {getattr(type, '__name__', 'type'): type}
    if length is not None:
        error = _format_error(
            error or 'should be as {} and with a length of {}!'.format(
                type, length
            )
        )
        return And(_type(type=type), _check_length(length), error=error)
    if not isinstance(type, (Use, Schema, And, Or)):
        type = Or(type, Use(type))
    error = _format_error(error or 'should be as {}!'.format(type))
    return _eval(type, error=error, usersyms=usersyms)


# noinspection PyUnusedLocal
def _index_dict(error=None, **kwargs):
    error = error or 'cannot be parsed as {}!'.format({int: float})
    error = _format_error(error)
    c = {Use(int): Use(float)}
    s = And(dict, c)

    def _f(x):
        return {k: v for k, v in enumerate(x, start=1)}

    _tuple = Or(tuple, Use(lambda x: tuple(np.asarray(x).ravel())))
    return Or(s, And(_dict(), c), And(_type(type=_tuple), Use(_f), c),
              error=error)


def _np_array2list(x):
    r = x.astype(object)
    r[np.isnan(x)] = None
    return r.tolist()


# noinspection PyUnusedLocal
def _np_array(dtype=None, error=None, read=True, ravel=True, **kwargs):
    dtype = dtype or float
    error = _format_error(
        error or 'cannot be parsed as np.array dtype={}!'.format(dtype)
    )
    if read:
        if ravel:
            c = Use(lambda x: np.asarray(x, dtype=dtype).ravel())
        else:
            c = Use(lambda x: np.asarray(x, dtype=dtype))
        return Or(And(str, _eval(
            c, usersyms={'np.array': np.array}
        )), c, And(_type(), c), Empty(), error=error)
    else:
        return And(_np_array(dtype=dtype), Use(_np_array2list), error=error)


def _check_np_array_positive(x):
    """

    :param x:
        Array.
    :type x: numpy.array
    :return:
    """
    # noinspection PyUnresolvedReferences
    return (x >= 0).all()


# noinspection PyUnusedLocal
def _np_array_positive(dtype=None, error=None, read=True,
                       check=_check_np_array_positive, ravel=True, **kwargs):
    dtype = dtype or float
    error = _format_error(
        error or 'cannot be parsed because it should be an ' \
                 'np.array dtype={} and positive!'.format(dtype)
    )
    if read:
        c = And(
            Use(lambda x: np.asarray(x, dtype=dtype)),
            Use(lambda x: x.ravel() if ravel else x),
            check
        )
        return Or(And(str, _eval(
            c, usersyms={'np.array': np.array}
        )), c, And(_type(), c), Empty(), error=error)
    else:
        return And(
            _np_array_positive(dtype=dtype), Use(lambda x: x.tolist()),
            error=error
        )


# noinspection PyUnusedLocal
def _alternator_current_model(error=None, read=True, **kwargs):
    error = _format_error(error)
    if read:
        from ..model.physical.electrics.motors.alternator.current import (
            AlternatorCurrentModel
        )
        return _type(type=AlternatorCurrentModel, error=error)
    return And(_alternator_current_model(), Use(lambda x: sh.NONE), error=error)


# noinspection PyUnusedLocal
def _service_battery_status_model(error=None, read=True, **kwargs):
    error = _format_error(error)
    if read:
        from ..model.physical.electrics.batteries.service.status import (
            BatteryStatusModel
        )
        return _type(type=BatteryStatusModel, error=error)
    return And(
        _service_battery_status_model(), Use(lambda x: sh.NONE), error=error
    )


# noinspection PyUnusedLocal
def _engine_temperature_regression_model(error=None, read=True, **kwargs):
    error = _format_error(error)
    if read:
        # noinspection PyProtectedMember
        from ..model.physical.engine._thermal import ThermalModel
        return _type(type=ThermalModel, error=error)
    return And(
        _engine_temperature_regression_model(), Use(lambda x: sh.NONE),
        error=error
    )


# noinspection PyUnusedLocal
def _fmep_model(error=None, read=True, **kwargs):
    error = _format_error(error)
    if read:
        from ..model.physical.engine.fc import FMEP
        return _type(type=FMEP, error=error)
    return And(_fmep_model(), Use(lambda x: sh.NONE), error=error)


# noinspection PyUnusedLocal
def _cmv(error=None, **kwargs):
    error = _format_error(error)
    from ..model.physical.gear_box.at_gear.cmv import CMV
    return _type(type=CMV, error=error)


# noinspection PyUnusedLocal
def _mvl(error=None, **kwargs):
    error = _format_error(error)
    from ..model.physical.gear_box.at_gear import MVL
    return _type(type=MVL, error=error)


# noinspection PyUnusedLocal
def _gspv(error=None, **kwargs):
    error = _format_error(error)
    from ..model.physical.gear_box.at_gear.gspv import GSPV
    return _type(type=GSPV, error=error)


# noinspection PyUnusedLocal
def _gsch(error=None, **kwargs):
    error = _format_error(error)
    from ..model.physical.gear_box.at_gear.gspv_ch import GSMColdHot
    return _type(type=GSMColdHot, error=error)


# noinspection PyUnusedLocal
def _dtc(error=None, read=True, **kwargs):
    error = _format_error(error)
    if read:
        from ..model.physical.gear_box.at_gear.dtgs import DTGS
        return _type(type=DTGS, error=error)
    return And(_dtc(), Use(lambda x: sh.NONE), error=error)


# noinspection PyUnusedLocal
def _cvt(error=None, read=True, **kwargs):
    error = _format_error(error)
    if read:
        from ..model.physical.gear_box.cvt import CVT
        return _type(type=CVT, error=error)
    return And(_dtc(), Use(lambda x: sh.NONE), error=error)


def _parameters2str(data):
    from lmfit import Parameters
    if isinstance(data, Parameters):
        return data.dumps(sort_keys=True)


def _str2parameters(data):
    if isinstance(data, str):
        from lmfit import Parameters
        p = Parameters()
        p.loads(data)
        return p
    return data


def _parameters(error=None, read=True):
    error = _format_error(error)
    if read:
        from lmfit import Parameters
        return And(Use(_str2parameters), _type(type=Parameters, error=error))
    else:
        return And(_parameters(), Use(_parameters2str), error=error)


# noinspection PyUnusedLocal
def _compare_str(s, **kwargs):
    return And(Use(str.lower), s.lower(), Use(lambda x: s))


# noinspection PyUnusedLocal
def _tyre_code(error=None, **kwargs):
    error = error or 'invalid tyre code!'
    error = _format_error(error)
    # noinspection PyProtectedMember
    from ..model.physical.wheels import (
        _re_tyre_code_iso, _re_tyre_code_numeric, _re_tyre_code_pax
    )
    return And(str, Or(
        _re_tyre_code_iso.match, _re_tyre_code_numeric.match,
        _re_tyre_code_pax.match
    ), error=error)


# noinspection PyUnusedLocal
def _tyre_dimensions(error=None, **kwargs):
    error = error or 'invalid format for tyre dimensions!'
    error = _format_error(error)
    # noinspection PyProtectedMember
    from ..model.physical.wheels import _format_tyre_dimensions
    return And(_dict(format=dict), Use(_format_tyre_dimensions), error=error)


def check_phases_separated(x):
    """
    >>> bags = [
        [3, 2, 3, 2, 4, 4, 1, 1], # INVALID!
        [3, 2, 2, 2, 4, 4, 1, 4], # INVALID!
        [3, 3, 2, 2, 4, 4, 1, 1], # valid
        ['P1', 'P3', 'P3', 'P2'], # valid
        [False, False],           # valid
        [],                       # valid
    ]
    >>> [check_phases_separated(x) for x in bags]
    [False, False, True, True, True, True]
    """
    if not len(x):
        return True

    x = np.asarray(x)
    # noinspection PyUnresolvedReferences
    deduped_count = 1 + (x[1:] != x[:-1]).sum()  # [3,3,1,1,3] --> len([3,1,3])

    return deduped_count == np.unique(x).size


# noinspection PyUnusedLocal
def _bag_phases(error=None, read=True, **kwargs):
    er = 'Phases must be separated!'
    error = _format_error(error)
    if read:
        return And(_np_array(read=read),
                   Schema(check_phases_separated, error=er), error=error)
    else:
        return And(_bag_phases(error), _np_array(read=False), error=error)


# noinspection PyUnusedLocal
def _file(error=None, **kwargs):
    er = 'Must be a file!'
    error = _format_error(error)
    return And(_string(), Schema(osp.isfile, error=er), error=error)


# noinspection PyUnusedLocal
def _dir(error=None, **kwargs):
    er = 'Must be a directory!'
    error = _format_error(error)

    def _fun(x):
        return not osp.exists(x) or osp.isdir(x)

    return And(_string(), Schema(_fun, error=er), error=error)


def _is_sorted(iterable, key=lambda a, b: a <= b):
    return all(key(a, b) for a, b in co2_utl.pairwise(iterable))


def _validator(schema, convert, default, k, v):
    if k in schema:
        default = schema[k]
    elif k.lower() in convert:
        k = convert[k.lower()]
        default = schema[k]
    return k, default.validate(v)


# noinspection PyUnresolvedReferences
@functools.lru_cache(None)
def define_data_validation(read=True):
    """
    Define data schema.

    :param read:
        Schema for reading?
    :type read: bool

    :return:
        Data schema.
    :rtype: function
    """
    cmv = _cmv(read=read)
    dtc = _dtc(read=read)
    cvt = _cvt(read=read)
    gspv = _gspv(read=read)
    gsch = _gsch(read=read)
    string = _string(read=read)
    positive = _positive(read=read)
    greater_than_zero = _positive(
        read=read, error='should be as <float> and greater than zero!',
        check=lambda x: x > 0
    )
    between_zero_and_one = _positive(
        read=read, error='should be as <float> and between zero and one!',
        check=lambda x: 0 <= x <= 1
    )
    greater_than_one = _positive(
        read=read, error='should be as <float> and greater than one!',
        check=lambda x: x >= 1
    )
    positive_int = _positive(type=int, read=read)
    greater_than_one_int = _positive(
        type=int, read=read, error='should be as <int> and greater than one!',
        check=lambda x: x >= 1
    )
    limits = _limits(read=read)
    index_dict = _index_dict(read=read)
    np_array = _np_array(read=read)
    np_array_sorted = _np_array_positive(
        read=read, error='cannot be parsed because it should be an '
                         'np.array dtype=<float> with ascending order!',
        check=_is_sorted
    )
    np_array_greater_than_minus_one = _np_array_positive(
        read=read, error='cannot be parsed because it should be an '
                         'np.array dtype=<float> and all values >= -1!',
        check=lambda x: (x >= -1).all()
    )
    np_array_bool = _np_array(dtype=bool, read=read)
    np_array_int = _np_array(dtype=int, read=read)
    _bool = _type(type=bool, read=read)
    function = _function(read=read)
    _tuple = Use(lambda x: tuple(np.asarray(x).ravel()))
    tuplefloat2 = _type(
        type=And(_tuple, (_type(float),)),
        length=2,
        read=read
    )
    _float = _type(type=float, read=read)
    _float = Or(_float, And(str, Use(lambda x: x.replace(',', '.')), _float))
    tuplefloat = _type(type=And(_tuple, (_type(float),)), read=read)
    dictstrdict = _dict(format={str: dict}, read=read)
    ordictstrdict = _ordict(format={str: dict}, read=read)
    parameters = _parameters(read=read)
    dictstrfloat = _dict(format={str: Use(float)}, read=read)
    dictarray = _dict(format={str: np_array}, read=read)
    tyre_code = _tyre_code(read=read)
    tyre_dimensions = _tyre_dimensions(read=read)
    maximum_velocity_range = _select(
        types=tuple(map(str, range(151))) + ('>150',), read=read
    )
    maximum_velocity_range = Or(
        And(Use(int), Use(str), maximum_velocity_range), maximum_velocity_range
    )
    convert = {
        'cvt': 'CVT',
        'cmv': 'CMV',
        'cmv_cold_hot': 'CMV_Cold_Hot',
        'dtgs': 'DTGS',
        'gspv': 'GSPV',
        'gspv_cold_hot': 'CMV_Cold_Hot',
        'mvl': 'MVL',
        'ki_factor': 'ki_multiplicative',
        'engine_max_speed_at_max_power': 'engine_speed_at_max_power',
        'battery_voltage': 'service_battery_nominal_voltage',
        'battery_capacity': 'service_battery_capacity',
        'state_of_charge_balance': 'service_battery_state_of_charge_balance',
        'state_of_charge_balance_window':
            'service_battery_state_of_charge_balance_window',
        'initial_state_of_charge': 'initial_service_battery_state_of_charge',
        'co2_emission_udc': 'co2_emission_UDC',
        'co2_emission_eudc': 'co2_emission_EUDC',
        'eco_mode': 'fuel_saving_at_strategy',
        'electric_load': 'service_battery_load',
        'alternator_statuses': 'service_battery_charging_statuses',
        'battery_currents': 'service_battery_currents',
        'state_of_charges': 'service_battery_state_of_charges',
        'obd_velocities': 'obd_velocities'
    }

    schema = {
        'CVT': cvt,
        'CMV': cmv,
        'CMV_Cold_Hot': gsch,
        'DTGS': dtc,
        'GSPV': gspv,
        'GSPV_Cold_Hot': gsch,
        'MVL': _mvl(read=read),
        'engine_n_cylinders': positive_int,
        'lock_up_tc_limits': tuplefloat2,
        'ki_multiplicative': greater_than_one,
        'ki_additive': positive,
        'drive_battery_technology': string,
        'drive_battery_n_cells': greater_than_one_int,
        'drive_battery_n_series_cells': greater_than_one_int,
        'drive_battery_n_parallel_cells': greater_than_one_int,
        'tyre_dimensions': tyre_dimensions,
        'tyre_code': tyre_code,
        'front_tyre_code': tyre_code,
        'rear_tyre_code': tyre_code,
        'wheel_drive': _select(
            types=('front', 'rear', 'front+rear'), read=read
        ),
        'wltp_base_model': _dict(format=dict, read=read),
        'fuel_type': _select(types=(
            'gasoline', 'diesel', 'LPG', 'NG', 'ethanol', 'biodiesel',
            'methanol', 'propane'), read=read),
        'obd_fuel_type_code': positive_int,
        'vehicle_category': _select(types='ABCDEFSMJ', read=read),
        'vehicle_body': _select(types=(
            'cabriolet', 'sedan', 'hatchback', 'stationwagon', 'suv/crossover',
            'mpv', 'coupé', 'bus', 'bestelwagen', 'pick-up'
        ), read=read),
        'tyre_class': _select(types=('C1', 'C2', 'C3'), read=read),
        'tyre_category': _select(types='ABCDEFG', read=read),
        'engine_fuel_lower_heating_value': positive,
        'fuel_carbon_content': positive,
        'engine_capacity': positive,
        'engine_stroke': positive,
        'engine_max_power': positive,
        'engine_speed_at_max_power': positive,
        'engine_max_speed': positive,
        'engine_max_torque': positive,
        'idle_engine_speed_median': positive,
        'engine_idle_fuel_consumption': greater_than_zero,
        'final_drive_ratio': positive,
        'r_dynamic': positive,
        'n_wheel': positive_int,
        'wheel_drive_load_fraction': between_zero_and_one,
        'static_friction': greater_than_zero,
        'tyre_state': _select(types=('new', 'worm'), read=read),
        'road_state': _select(
            types=('dry', 'wet', 'rainfall', 'puddles', 'ice'), read=read
        ),
        'wltp_class': _select(types=('class1', 'class2', 'class3a', 'class3b'),
                              read=read),
        'downscale_phases': tuplefloat,
        'electrical_hybridization_degree': _select(
            types=('none', 'mild', 'full', 'plugin', 'electric'), read=read
        ),
        'gear_box_type': _select(
            types=('manual', 'automatic', 'cvt', 'planetary', 'none'), read=read
        ),
        'initial_temperature': _float,
        'ignition_type': _select(types=('positive', 'compression'), read=read),
        'start_stop_activation_time': positive,
        'alternator_nominal_voltage': positive,
        'service_battery_nominal_voltage': positive,
        'service_battery_capacity': positive,
        'service_battery_state_of_charge_balance': limits,
        'service_battery_state_of_charge_balance_window': limits,
        'initial_service_battery_state_of_charge': limits,
        'idle_engine_speed_std': positive,
        'alternator_nominal_power': positive,
        'alternator_efficiency': _limits(limits=(0, 1), read=read),
        'time_cold_hot_transition': positive,
        'co2_params': dictstrfloat,
        'willans_factors': dictstrfloat,
        'phases_willans_factors': _type(
            type=And(Use(tuple), (dictstrfloat,)), read=read),
        'optimal_efficiency': dictarray,
        'velocity_speed_ratios': index_dict,
        'gear_box_ratios': index_dict,
        'final_drive_ratios': index_dict,
        'speed_velocity_ratios': index_dict,
        'full_load_speeds': np_array_sorted,
        'full_load_torques': np_array,
        'full_load_powers': np_array,

        'capped_velocity': positive,
        'has_capped_velocity': _bool,
        'maximum_velocity_range': maximum_velocity_range,
        'maximum_velocity': positive,
        'vehicle_mass_running_order': positive,
        'vehicle_mass': positive,
        'f2': positive,
        'correct_f0': _bool,

        'co2_emission_low': positive,
        'co2_emission_medium': positive,
        'co2_emission_high': positive,
        'co2_emission_extra_high': positive,

        'co2_emission_UDC': positive,
        'co2_emission_EUDC': positive,
        'co2_emission_value': positive,
        'declared_co2_emission_value': positive,
        'n_dyno_axes': positive_int,
        'n_wheel_drive': positive_int,
        'rcb_correction': _bool,
        'speed_distance_correction': _bool,
        'engine_is_turbo': _bool,
        'has_start_stop': _bool,
        'has_gear_box_thermal_management': _bool,
        'has_energy_recuperation': _bool,
        'is_hybrid': _bool,
        'has_roof_box': _bool,
        'has_periodically_regenerating_systems': _bool,
        'engine_has_variable_valve_actuation': _bool,
        'has_thermal_management': _bool,
        'engine_has_direct_injection': _bool,
        'has_lean_burn': _bool,
        'engine_has_cylinder_deactivation': _bool,
        'has_exhausted_gas_recirculation': _bool,
        'has_particle_filter': _bool,
        'has_selective_catalytic_reduction': _bool,
        'has_nox_storage_catalyst': _bool,
        'has_torque_converter': _bool,
        'is_cycle_hot': _bool,
        'is_serial': _bool,
        'use_dt_gear_shifting': _bool,
        'fuel_saving_at_strategy': _bool,
        'correct_start_stop_with_gears': _bool,
        'enable_phases_willans': _bool,
        'enable_willans': _bool,
        'has_engine_idle_coasting': _bool,
        'has_engine_off_coasting': _bool,
        'fuel_map': dictarray,
        'transition_cycle_index': positive_int,
        'alternator_charging_currents': tuplefloat2,
        'relative_electric_energy_change': tuplefloat,
        'alternator_current_model': _alternator_current_model(read=read),
        'dcdc_current_model': _alternator_current_model(read=read),
        'service_battery_status_model': _service_battery_status_model(
            read=read),
        'clutch_speed_model': function,
        'co2_emissions_model': function,
        'co2_error_function_on_emissions': function,
        'co2_error_function_on_phases': function,
        'motor_p0_electric_power_loss_function': function,
        'motor_p1_electric_power_loss_function': function,
        'motor_p2_electric_power_loss_function': function,
        'motor_p3_front_electric_power_loss_function': function,
        'motor_p3_rear_electric_power_loss_function': function,
        'motor_p4_front_electric_power_loss_function': function,
        'motor_p4_rear_electric_power_loss_function': function,
        'after_treatment_speed_model': function,
        'after_treatment_power_model': function,
        'clutch_window': tuplefloat2,
        'co2_params_calibrated': parameters,
        'co2_params_initial_guess': parameters,
        'drive_battery_technology_type': string,
        'cycle_type': string,
        'cycle_name': string,
        'specific_gear_shifting': string,
        'calibration_status': _type(
            type=And(Use(list), [(bool, Or(parameters, None))]), length=4,
            read=read
        ),
        'service_battery_load': tuplefloat2,
        'engine_thermostat_temperature_window': tuplefloat2,
        'engine_temperature_regression_model':
            _engine_temperature_regression_model(read=read),
        'engine_type': string,
        'input_type': string,
        'starter_model': function,
        'drive_battery_model': function,
        'motor_p0_maximum_power_function': function,
        'motor_p1_maximum_power_function': function,
        'motor_p2_planetary_maximum_power_function': function,
        'start_stop_hybrid_params': dictstrfloat,
        'full_load_curve': function,
        'fmep_model': _fmep_model(read=read),
        'gear_box_efficiency_constants': dictstrdict,
        'gear_box_efficiency_parameters_cold_hot': dictstrdict,
        'scores': dictstrdict,
        'param_selections': dictstrdict,
        'model_selections': dictstrdict,
        'score_by_model': dictstrdict,
        'at_scores': ordictstrdict,

        'fuel_density': positive,
        'idle_engine_speed': tuplefloat2,
        'k1': positive_int,
        'k2': positive_int,
        'k5': positive_int,
        'max_gear': positive_int,
        'hybrid_modes': np_array_int,
        'road_loads': _type(type=And(_tuple, (_type(float),)),
                            length=3,
                            read=read),
        'start_stop_model': function,
        'gear_box_temperature_references': tuplefloat2,
        'torque_converter_speed_model': function,
        'phases_co2_emissions': tuplefloat,
        'bag_phases': _bag_phases(read=read),
        'phases_integration_times':
            _type(type=And(Use(tuple), (And(Use(tuple), (_type(float),)),)),
                  read=read),
        'active_cylinder_ratios': tuplefloat,
        'extended_phases_co2_emissions': tuplefloat,
        'extended_phases_integration_times':
            _type(type=And(Use(tuple), (And(Use(tuple), (_type(float),)),)),
                  read=read),
        'extended_integration_times': tuplefloat,
        'phases_fuel_consumptions': tuplefloat,
        'co2_rescaling_scores': tuplefloat,
        'accelerations': np_array,
        'alternator_currents': np_array,
        'active_cylinders': np_array,
        'alternator_powers': np_array,
        'service_battery_charging_statuses': np_array_int,
        'auxiliaries_power_losses': np_array,
        'auxiliaries_torque_loss_factors': tuplefloat,
        'auxiliaries_torque_losses': np_array,
        'service_battery_currents': np_array,
        'clutch_tc_powers': np_array,
        'clutch_tc_speeds_delta': np_array,
        'co2_emissions': np_array,
        'after_treatment_speeds_delta': np_array,
        'engine_coolant_temperatures': np_array,
        'engine_powers_out': np_array,
        'engine_speeds_out': np_array,
        'engine_speeds_out_hot': np_array,
        'engine_starts': np_array_bool,
        'co2_normalization_references': np_array,
        'final_drive_powers_in': np_array,
        'final_drive_speeds_in': np_array,
        'final_drive_torques_in': np_array,
        'fuel_consumptions': np_array,
        'fuel_consumptions_liters': np_array,
        'gear_box_efficiencies': np_array,
        'gear_box_powers_in': np_array,
        'gear_box_speeds_in': np_array,
        'gear_box_temperatures': np_array,
        'gear_box_torque_losses': np_array,
        'gear_box_torques_in': np_array,
        'gear_shifts': np_array_bool,
        'gears': np_array_int,
        'identified_co2_emissions': np_array,
        'motive_powers': np_array,
        'on_engine': np_array_bool,
        'clutch_phases': np_array_bool,
        'after_treatment_warm_up_phases': np_array_bool,
        'on_idle': np_array_bool,
        'service_battery_state_of_charges': np_array,
        'times': np_array_sorted,
        'velocities': np_array_greater_than_minus_one,
        'obd_velocities': np_array_greater_than_minus_one,
        'wheel_powers': np_array,
        'wheel_speeds': np_array,
        'wheel_torques': np_array,
    }
    try:
        from co2mpas_driver.co2mpas import plugin_schema

        schema, convert = plugin_schema(schema, convert)
    except ImportError:
        pass

    schema = {k: Or(Empty(), v) for k, v in schema.items()}
    schema[sh.NONE] = Or(_float, np_array)

    if not read:
        def _f(x):
            return x is sh.NONE

        schema = {k: And(v, Or(_f, Use(str))) for k, v in schema.items()}

    return functools.partial(_validator, schema, convert, schema.pop(sh.NONE))


def _input_version(error=None, read=True, **kwargs):
    def _check_data_version(input_version):
        from co2mpas import __file_version__ as exp_ver
        exp_vinfo = tuple(exp_ver.split('.'))
        got_vinfo = tuple(input_version.split('.'))

        if not read or got_vinfo[:2] == exp_vinfo[:2]:
            return True

        if got_vinfo[:1] != exp_vinfo[:1]:
            msg = "Input-file version %s is incompatible with expected %s.\n" \
                  "  More failures may happen."

        elif got_vinfo[:2] > exp_vinfo[:2]:
            msg = "Input-file version %s comes from the (incompatible) " \
                  "future (> %s).\n  More failures may happen."
        else:
            msg = "Input-file version %s is old (< %s)." \
                  "\n  You may need to update it, to use new fields."
        log.warning(msg, input_version, exp_ver)
        return True

    return And(_string(**kwargs), _check_data_version, error=error)


@functools.lru_cache(None)
def define_flags_validation(read=True):
    """
    Define flag schema.

    :param read:
        Schema for reading?
    :type read: bool

    :return:
        Flag schema.
    :rtype: function
    """
    string = _string(read=read)
    isfile = _file(read=read)
    isdir = _dir(read=read)
    _bool = _type(type=bool, read=read)

    schema = {
        'input_version': _input_version(read=read),
        'model_conf': isfile,
        'encryption_keys': string,
        'encryption_keys_passwords': string,
        'sign_key': string,

        'hard_validation': _bool,
        'enable_selector': _bool,
        'declaration_mode': _bool,
        'only_summary': _bool,
        'augmented_summary': _bool,
        'type_approval_mode': _bool,

        'output_template': isfile,
        'output_folder': isdir,
    }

    schema = {k: Or(Empty(), v) for k, v in schema.items()}

    if not read:
        def _f(x):
            return x is sh.NONE

        schema = {k: And(v, Or(_f, Use(str))) for k, v in schema.items()}

    return functools.partial(_validator, schema, {k: k for k in schema}, None)
